/*
 * libvcd.h
 * Coypright (C) 2000 Herbert Valerio Riedel <hvr@gnu.org>
 * Coypright (C) 2000 Jens B. Jorgensen <jbj1@ultraemail.net>
 * Coypright (C) 2005 Rocky Bernstein <rocky@panix.com>
 * Copyright (C) 2006 - 2008 Josh Sutherland <OneThirty8@aol.com>
 * This file is part of VC/DC, a free MPEG-2 video player.
 *
 * VC/DC is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * VC/DC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
/* Most of this stuff is from vcdimager.  Some pieces are from libcdio, by the
   same authors as vcdimager.  I'm only taking what I absolutely need, since
   vcdimager seems to be way more complicated than this stuff needs to be for
   Dreamcast.  We only have one CD drive, so we're only going to be dealing with
   one VCD at a time.  Our code does _NOT_ need to be reentrant, and it's easier
   to to wrap my mind around if it is not, so although I am probably reinventing
   the wheel here, our VCD library should be a hell of a lot easier for somebody
   to navigate through to figure out what is going on.
   
   All I can really take credit for in this file will be down at the bottom of the
   file, where there are some macros ripped from libcdio.  I used that code to make
   the info in the ENTRIES.VCD file easier for me to play with, and wrote a macro to
   handle that conversion.
   
   -JHS, 2008.
*/
#ifndef LIBVCD_H
#define LIBVCD_H
#include <kos.h>
#define EMPTY_ARRAY_SIZE
#define ISO_BLOCKSIZE 2048
/* random note: most stuff is big endian here */
#define GNUC_PACKED __attribute__((packed))
#define BITFIELD_LSBF

#define ENTRIES_ID_VCD  "ENTRYVCD"
#define ENTRIES_ID_VCD3 "ENTRYSVD"
#define ENTRIES_ID_SVCD "ENTRYVCD" /* not ENTRYSVD! */

#define ENTRIES_VERSION_VCD   0x01
#define ENTRIES_SPTAG_VCD     0x00   

#define ENTRIES_VERSION_VCD11 0x01
#define ENTRIES_SPTAG_VCD11   0x00   

#define ENTRIES_VERSION_VCD2  0x02
#define ENTRIES_SPTAG_VCD2    0x00

#define ENTRIES_VERSION_SVCD  0x01
#define ENTRIES_SPTAG_SVCD    0x00

#define ENTRIES_VERSION_HQVCD 0x01
#define ENTRIES_SPTAG_HQVCD   0x00

/* I actually don't use this next chunk of sector addresses--they're off by 150(LBA and LSN differ by 150).
  INFO.VCD is at sector 300, ENTRIES.VCD is sector 301, etc.  But the info here is still useful,
  so I left it to refer to. */
#define INFO_VCD_SECTOR    150
#define ENTRIES_VCD_SECTOR 151
#define LOT_VCD_SECTOR     152
#define LOT_VCD_SIZE       32
#define PSD_VCD_SECTOR     (LOT_VCD_SECTOR+LOT_VCD_SIZE)

#define MAX_SEGMENTS       1980
#define MAX_ENTRIES        500
#define MAX_SEQ_ENTRIES    99

/* these are used for SVCDs only */
#define TRACKS_SVD_SECTOR  (PSD_VCD_SECTOR+1)
#define SEARCH_DAT_SECTOR  (TRACKS_SVD_SECTOR+1)

/* Maximum index of optional LOT.VCD (the List ID Offset Table.) */
#define LOT_VCD_OFFSETS ((1 << 15)-1)

#if defined(__GNUC__)
  /* this is strict ISO C99 which allows only 'unsigned int', 'signed
     int' and '_Bool' explicitly as bit-field type */
  typedef unsigned int bitfield_t;
  //typedef unsigned char bitfield_t;
#endif
#define PRAGMA_BEGIN_PACKED
#define PRAGMA_END_PACKED
PRAGMA_BEGIN_PACKED
struct msf_s {
    uint8 m, s, f; /* BCD encoded! */
  } GNUC_PACKED;
  
typedef struct msf_s msf_t;

typedef struct _EntriesVcd_tag {
  char ID[8];                             /**< "ENTRYVCD" or "ENTRYSVD" */
  uint8 version;                        /**< 0x02 --- VCD2.0
                                               0x01 --- SVCD, should be
                                               same as version in
                                               INFO.SVD */
  uint8 sys_prof_tag;                   /**< 0x01 if VCD1.1
                                                0x00 else */
  uint16 entry_count;                   /**< 1 <= tracks <= 500 */
  struct {     /* all fields are BCD */
    uint8 n; /* cd track no 2 <= n <= 99 */
    msf_t msf;
  } GNUC_PACKED entry[MAX_ENTRIES];
  uint8 reserved2[36];                  /* RESERVED, must be 0x00 */
} GNUC_PACKED _EntriesVcd_t; /* sector 00:04:01 */

#define EntriesVcd_t_SIZEOF ISO_BLOCKSIZE


#define INFO_ID_VCD   "VIDEO_CD"
#define INFO_ID_SVCD  "SUPERVCD"
#define INFO_ID_HQVCD "HQ-VCD  "

#define INFO_VERSION_VCD   0x01
#define INFO_SPTAG_VCD     0x00   

#define INFO_VERSION_VCD11 0x01
#define INFO_SPTAG_VCD11   0x01   

#define INFO_VERSION_VCD2  0x02
#define INFO_SPTAG_VCD2    0x00   

#define INFO_VERSION_SVCD  0x01
#define INFO_SPTAG_SVCD    0x00

#define INFO_VERSION_HQVCD 0x01
#define INFO_SPTAG_HQVCD   0x01

#define INFO_OFFSET_MULT   0x08

/** InfoStatusFlags: this one-byte field describes certain
    characteristics of the disc */
    
typedef struct {
#if defined(BITFIELD_LSBF)
  bool       reserved1 : 1;                   /* Reserved, must be zero */
  bitfield_t restriction : 2;                 /* restriction, eg. "unsuitable
                                              for kids":
                                              0x0 ==> unrestricted,
                                              0x1 ==> restricted category 1,
                                              0x2 ==> restricted category 2,
                                              0x3 ==> restricted category 3 */
  bool       special_info : 1;                /**< Special Information is 
                                                   encoded 
                                                   in the pictures */
  bool       user_data_cc : 1;                /**< MPEG User Data is used
                                                   for Closed Caption */
  bool       use_lid2 : 1;                    /**<  If == 1 and the PSD is
                                                 interpreted and the next
                                                 disc has the same album
                                                 id then start the next
                                                 disc at List ID #2,
                                                 otherwise List ID #1 */ 
  bool       use_track3 : 1;                  /**< If == 1 and the PSD is
                                                 not interpreted  and
                                                 next disc has same album
                                                 id, then start next disc
                                                 with track 3, otherwise
                                                 start with track 2 */ 
  bool       pbc_x : 1;                       /**< extended PBC available */
#else
  bool       pbc_x : 1;
  bool       use_track3 : 1;
  bool       use_lid2 : 1;
  bool       user_data_cc : 1;
  bool       special_info : 1;
  bitfield_t restriction : 2;
  bool       reserved1 : 1;
#endif
} GNUC_PACKED InfoStatusFlags_t;

#define InfoStatusFlags_t_SIZEOF 1

enum {
  VCD_FILES_VIDEO_NOSTREAM = 0,
  VCD_FILES_VIDEO_NTSC_STILL = 1,
  VCD_FILES_VIDEO_NTSC_STILL2 = 2,
  VCD_FILES_VIDEO_NTSC_MOTION = 3,
  VCD_FILES_VIDEO_PAL_STILL = 5,
  VCD_FILES_VIDEO_PAL_STILL2 = 6,
  VCD_FILES_VIDEO_PAL_MOTION = 7
};

typedef struct 
{
#if defined(BITFIELD_LSBF)
  bitfield_t audio_type : 2;                /**< Audio characteristics:
                                              0x0 - No MPEG audio stream
                                              0x1 - One MPEG1 or MPEG2 audio
                                                    stream without extension
                                              0x2 - Two MPEG1 or MPEG2 audio
                                                    streams without extension
                                              0x3 - One MPEG2 multi-channel 
                                                    audio stream w/ extension*/
  bitfield_t video_type : 3;                /**< Video characteristics:
                                              0x0 - No MPEG video data
                                              0x1 - NTSC still picture
                                              0x2 - Reserved (NTSC hires?)
                                              0x3 - NTSC motion picture
                                              0x4 - Reserved
                                              0x5 - PAL still picture
                                              0x6 - Reserved (PAL hires?)
                                              0x7 - PAL motion picture */
  bool       item_cont : 1;                 /**< Indicates segment is 
                                                 continuation
                                              0x0 - 1st or only segment of item
                                              0x1 - 2nd or later
                                                    segment of item */
  bitfield_t ogt : 2;                       /**< 0x0 - no OGT substream 
                                                 0x1 - sub-stream 0 available
                                                 0x2 - sub-stream 0 & 1 
                                                       available
                                                 0x3 - all OGT sub-substreams 
                                                       available */
#else
  bitfield_t ogt : 2;
  bool       item_cont : 1;
  bitfield_t video_type : 3;
  bitfield_t audio_type : 2;
#endif
} GNUC_PACKED InfoSpiContents_t;

#define InfoSpiContents_t_SIZEOF 1

typedef struct _InfoVcd_tag {
  char   ID[8];              /**< const "VIDEO_CD" for
                                  VCD, "SUPERVCD" or
                                  "HQ-VCD  " for SVCD */
  uint8 version;           /**< 0x02 -- VCD2.0,
                                  0x01 for SVCD and VCD1.x */ 
  uint8 sys_prof_tag;      /**< System Profile Tag, used
                                  to define the set of
                                  mandatory parts to be
                                  applied for compatibility;
                                  0x00 for "SUPERVCD",
                                  0x01 for "HQ-VCD  ",
                                  0x0n for VCDx.n */ 
  char album_desc[16];       /**< album identification/desc. */
  uint16 vol_count;        /**< number of volumes in album */
  uint16 vol_id;           /**< number id of this volume in album */
  uint8  pal_flags[13];    /**< bitset of 98 PAL(=set)/NTSC flags */
  InfoStatusFlags_t flags;   /**< status flags bit field */
  uint32 psd_size;         /**< size of PSD.VCD file */
  msf_t    first_seg_addr;   /**< first segment addresses,
                                  coded BCD The location
                                  of the first sector of
                                  the Segment Play Item
                                  Area, in the form
                                  mm:ss:00. Must be
                                  00:00:00 if the PSD size
                                  is 0. */
  uint8  offset_mult;      /**< offset multiplier, must be 8 */
  uint16 lot_entries;      /**< offsets in lot */
  uint16 item_count;       /**< segments used for segmentitems */
  
  InfoSpiContents_t spi_contents[MAX_SEGMENTS]; /**< The next 1980 bytes
                                                   contain one byte for each 
                                                   possible segment play item. 
                                                   Each byte indicates
                                                   contents. */

  uint16 playing_time[5];  /**< in seconds */
  char reserved[2];          /**< Reserved, must be zero */
} GNUC_PACKED _InfoVcd_t;

#define InfoVcd_t_SIZEOF ISO_BLOCKSIZE

/** LOT.VCD
   This optional file is only necessary if the PSD size is not zero.
   This List ID Offset Table allows you to start playing the PSD from
   lists other than the default List ID number. This table has a fixed length
   of 32 sectors and maps List ID numbers into List Offsets. It's got
   an entry for each List ID Number with the 16-bit offset. Note that
   List ID 1 has an offset of 0x0000. All unused or non-user-accessible
   entries must be 0xffff. */

#define LOT_VCD_OFFSETS ((1 << 15)-1)

typedef struct _LotVcd_tag {
  uint16 reserved;  /* Reserved, must be zero */
  uint16 offset[LOT_VCD_OFFSETS];  /* offset given in 8 byte units */
} GNUC_PACKED _LotVcd_t;

#define LotVcd_t_SIZEOF (32*ISO_BLOCKSIZE)

/** PSD.VCD
   The PSD controls the "user interaction" mode which can be used to make
   menus, etc. The PSD contains a set of Lists. Each List defines a set of
   Items which are played in sequence. An Item can be an mpeg track (in whole
   or part) or a Segment Play Item which can subsequently be mpeg video
   with or without audio, one more more mpeg still pictures (with or without
   audio) or mpeg audio only.

   The Selection List defines the action to be taken in response to a set
   of defined user actions: Next, Previous, Default Select, Numeric, Return.

   The End List terminates the control flow or switches to the next
   disc volume.

   Each list has a unique list id number. The first must be 1, the others can
   be anything (up to 32767).

   References to PSD list addresses are expressed as an offset into the PSD
   file. The offset indicated in the file must be multiplied by the Offset
   Multiplier found in the info file (although this seems to always have to
   be 8). Unused areas are filled with zeros. List ID 1 starts at offset 0.
*/

/* ...difficult to represent as monolithic C struct... */

typedef enum {
  PSD_TYPE_PLAY_LIST = 0x10,        /* Play List */
  PSD_TYPE_SELECTION_LIST = 0x18,   /* Selection List (+Ext. for SVCD) */
  PSD_TYPE_EXT_SELECTION_LIST = 0x1a, /* Extended Selection List (VCD2.0) */
  PSD_TYPE_END_LIST = 0x1f,         /* End List */
  PSD_TYPE_COMMAND_LIST = 0x20      /* Command List */
} psd_descriptor_types;

typedef struct {
  uint8  type;        /* PSD_TYPE_END_LIST */
  uint8  next_disc;   /* 0x00 to stop PBC or 0xnn to switch to disc no nn */
  uint16 change_pic;  /* 0 or 1000..2979, should be still image */
  uint8  reserved[4]; /* padded with 0x00 */
} GNUC_PACKED PsdEndListDescriptor_t;

#define PsdEndListDescriptor_t_SIZEOF 8

/** (0,0) == upper left; (255,255) == lower right.
      Setting all to zero disables the area. */
typedef struct
  {
    uint8 x1; /**< upper left */
    uint8 y1; /**< upper left */
    uint8 x2; /**< lower right */
    uint8 y2; /**< lower right */
} GNUC_PACKED psd_area_t;

#define struct_psd_area_t_SIZEOF 4
  
#define PSD_OFS_DISABLED         0xffff /* Disables any particular offset. */
#define PSD_OFS_MULTI_DEF        0xfffe /* 'Default' points to different places depending on where in play_item. */
#define PSD_OFS_MULTI_DEF_NO_NUM 0xfffd /* Turns off buttons 1-99. */
#define JUMP_DELAYED 0x80		/* Jump immediately, or wait until the item is done playing? */

typedef struct {
#if defined(BITFIELD_LSBF)
  bool       SelectionAreaFlag : 1;
  bool       CommandListFlag : 1;
  bitfield_t reserved : 6;
#else
  bitfield_t reserved : 6;
  bool       CommandListFlag : 1;
  bool       SelectionAreaFlag : 1;
#endif  
} GNUC_PACKED PsdSelectionListFlags_t;

#define PsdSelectionListFlags_t_SIZEOF 1

typedef struct _PsdSelectionListDescriptor_tag {
  uint8 type;
  PsdSelectionListFlags_t flags;
  uint8 nos;
  uint8 bsn;
  uint16 lid;
  uint16 prev_ofs;
  uint16 next_ofs;
  uint16 return_ofs;
  uint16 default_ofs;
  uint16 timeout_ofs;
  uint8 totime;
  uint8 loop;
  uint16 itemid;
  uint16 ofs[EMPTY_ARRAY_SIZE]; /* variable length */
  /* PsdSelectionListDescriptorExtended */
} GNUC_PACKED _PsdSelectionListDescriptor_t;

#define PsdSelectionListDescriptor_t_SIZEOF 20

typedef struct {
  psd_area_t prev_area;
  psd_area_t next_area;
  psd_area_t return_area;
  psd_area_t default_area;
  psd_area_t area[EMPTY_ARRAY_SIZE]; /* variable length */
} GNUC_PACKED PsdSelectionListDescriptorExtended_t;

#define PsdSelectionListDescriptorExtended_t_SIZEOF 16

/* SVCD only, and not used by our player.  The list of offsets below maybe maps to numeric
   keys on a VCD player's remote? I don't know what we use the command list for.  Unless
   I missed something, both xine plugins ignore this, so that's what we're doing too. */
typedef struct {
  uint8 type;
  uint16 command_count;
  uint16 lid;
  uint16 command[EMPTY_ARRAY_SIZE]; /* variable length */
} GNUC_PACKED PsdCommandListDescriptor_t;

#define PsdCommandListDescriptor_t_SIZEOF 5

typedef struct  _PsdPlayListDescriptor_tag {
  uint8 type;
  uint8 noi;  /**< number of items */
  uint16 lid; /**< list id: high-bit means this list is rejected in
                     the LOT (also, can't use 0) */
  uint16 prev_ofs; /**< previous list offset (0xffff disables) */
  uint16 next_ofs; /**< next list offset (0xffff disables) */
  uint16 return_ofs; /**< return list offset (0xffff disables) */
  uint16 ptime; /**< play time in 1/15 s, 0x0000 meaning full item */
  uint8  wtime; /**< delay after, in seconds, if 1 <= wtime <= 60 wait
                       is wtime else if 61 <= wtime <= 254 wait is
                       (wtime-60) * 10 + 60 else wtime == 255 wait is
                       infinite  */
  uint8  atime; /**< auto pause wait time calculated same as wtime,
                       used for each item in list if the auto pause flag
                       in a sector is true */
  uint16 itemid[EMPTY_ARRAY_SIZE]; /**< item number
                                           0 <= n <= 1      - play nothing
                                           2 <= n <= 99     - play track n
                                         100 <= n <= 599    - play entry
                                                       (n - 99) from entries
                                                       table to end of track
                                         600 <= n <= 999    - reserved
                                        1000 <= n <= 2979   - play segment
                                                       play item (n - 999)
                                        2980 <= n <= 0xffff - reserved */
} GNUC_PACKED _PsdPlayListDescriptor_t;

  /* For backwards compatibility. Don't use PsdListDescriptor. */
#define _PsdPlayListDescriptor _PsdPlayListDescriptor_t
#define PsdPlayListDescriptor_t_SIZEOF 14

/* SCANDATA stuff will allow us to play in reverse and nifty stuff like that! */

/** SCANDATA.DAT for VCD 2.0 (ONLY! SVCD and HQVCD are different!) */

#define SCANDATA_FILE_ID "SCAN_VCD"
#define SCANDATA_VERSION_VCD2 0x02
#define SCANDATA_VERSION_SVCD 0x01

typedef struct {
  char file_id[sizeof(SCANDATA_FILE_ID)-1]; /**< == "SCAN_VCD" without final
                                                 NULL byte */
  uint8 version;                          /**< = 0x02 */
  uint8 reserved;                         /**< Reserved, must be zero */
  uint16 scan_points;                     /**< the number of scan points */
  msf_t points[EMPTY_ARRAY_SIZE];           /**< actual scan points 
                                               points[time(iframe)/0.5] */
} GNUC_PACKED ScandataDat_v2_t;

#define ScandataDat_v2_t_SIZEOF 12

/** SVCD/SEARCH.DAT
   This file defines where the scan points are. It covers all mpeg tracks
   together. A scan point at time T is the nearest I-picture in the MPEG
   stream to the given time T. Scan points are given at every half-second
   for the entire duration of the disc. */

#define SEARCH_FILE_ID        "SEARCHSV"
#define SEARCH_VERSION        0x01
#define SEARCH_TIME_INTERVAL  0x01

typedef struct {
  char file_id[sizeof(SEARCH_FILE_ID)-1]; /**< == "SEARCHSV" without final 
                                               NULL byte */
  uint8 version;  /**< = 0x01 */
  uint8 reserved; /**< Reserved, must be zero */
  uint16 scan_points; /**< the number of scan points */
  uint8 time_interval; /**< The interval of time in
                              between scan points, in units
                              of 0.5 seconds, must be 0x01 */
  msf_t points[EMPTY_ARRAY_SIZE]; /**< The series of scan points */
} GNUC_PACKED SearchDat_t;

#define SearchDat_t_SIZEOF 13

/** SCANDATA.DAT for SVCD
   This file fulfills much the same purpose of the SEARCH.DAT file except
   that this file is mandatory only if the System Profile Tag of the
   INFO.SVD file is 0x01 (HQ-VCD) and also that it contains sector addresses
   also for each video Segment Play Items in addition to the regular MPEG
   tracks. */

typedef struct {
  char file_id[sizeof(SCANDATA_FILE_ID)-1]; /**< == "SCAN_VCD" without final
                                                 NULL byte */
  uint8 version;                          /**<  = 0x01 */
  uint8 reserved;                         /**< Reserved, must be zero */
  uint16 scandata_count;                  /**< number of 3-byte entries in
                                                 the table */
  uint16 track_count;                     /**< number of MPEG tracks on 
                                               disc */
  uint16 spi_count;                       /**<  number of consecutively 
                                               recorded play item segments 
                                               (as opposed to the number of 
                                               segment play items). */
  msf_t cum_playtimes[EMPTY_ARRAY_SIZE];    /**<  cumulative playing
                                               time up to track
                                               N. Track time just wraps
                                               at 99:59:74 */
} GNUC_PACKED ScandataDat1_t;

#define ScandataDat1_t_SIZEOF 16

typedef struct {
  /* ScandataDat head; */
  uint16 spi_indexes[1]; /* should be [], but C doesn't allow that;
                              Indexes into the following scandata
                              table */
} GNUC_PACKED ScandataDat2_t;

#define ScandataDat2_t_SIZEOF sizeof(uint16_t)

typedef struct {
  /* ScandataDat2 head; */
  uint16 mpegtrack_start_index; /* Index into the
                                     following scandata table
                                     where the MPEG track
                                     scan points start */
  
  /* The scandata table starts here */
  struct {
    uint8 track_num;   /* Track number as in TOC */
    uint16 table_offset;   /* Index into scandata table */
  } GNUC_PACKED mpeg_track_offsets[EMPTY_ARRAY_SIZE];
} GNUC_PACKED ScandataDat3_t;

#define ScandataDat3_t_SIZEOF 2

typedef struct {
  /* ScandataDat3 head; */
  msf_t scandata_table[1]; /* should be [] but C99 doesn't allow that */
} GNUC_PACKED ScandataDat4_t;

#define ScandataDat4_t_SIZEOF msf_t_SIZEOF

/* My simplified struct for entries... We don't really care about a lot of
   this stuff but entry count and lba are useful.  MSF isn't as useful, and
   the rest should be the same as what we got when we opened INFO.VCD and if
   not, it's probably not the end of the world.
   These things are all either Big Endian or BCD-encoded, so we need to
   convert them, and that's what this struct is for. */
typedef struct _EntriesVcd_simple {
  uint16 entry_count;     /**< 1 <= tracks <= 500 */
  int track[MAX_ENTRIES];
  int ent_lba[MAX_ENTRIES];
} GNUC_PACKED _EntriesVcd_sim;


/* End of PBC stuff... */
/* Some macros from libcdio to make life simpler... */
#define UINT16_SWAP_LE_BE_C(val) ((uint16) ( \
    (((uint16) (val) & (uint16) 0x00ffU) << 8) | \
    (((uint16) (val) & (uint16) 0xff00U) >> 8)))

#define UINT32_SWAP_LE_BE_C(val) ((uint32) ( \
    (((uint32) (val) & (uint32) 0x000000ffU) << 24) | \
    (((uint32) (val) & (uint32) 0x0000ff00U) <<  8) | \
    (((uint32) (val) & (uint32) 0x00ff0000U) >>  8) | \
    (((uint32) (val) & (uint32) 0xff000000U) >> 24)))
    
/*Here's a macro to handle the conversion.  IN_ENT is _EntriesVcd_t,
   OUT_ENT is _EntriesVcd_sim. */
#define simplify_entries(IN_ENT, OUT_ENT)\
{ 	\
  uint8 p;  \
  int xyz; \
  OUT_ENT->entry_count = UINT16_SWAP_LE_BE_C(IN_ENT->entry_count); \
  for (xyz = 0; xyz <OUT_ENT->entry_count; xyz++) \
  {	\
   p = IN_ENT->entry[xyz].n;\
   OUT_ENT->track[xyz] = ((0xf & p)+(10*(p >> 4)));\
   p = IN_ENT->entry[xyz].msf.m; \
   OUT_ENT->ent_lba[xyz] = ((0xf & p)+(10*(p >> 4)));\
   OUT_ENT->ent_lba[xyz] *= 60; \
   p = IN_ENT->entry[xyz].msf.s; \
   OUT_ENT->ent_lba[xyz] += ((0xf & p)+(10*(p >> 4)));\
   OUT_ENT->ent_lba[xyz] *= 75; \
   p = IN_ENT->entry[xyz].msf.f; \
   OUT_ENT->ent_lba[xyz] += ((0xf & p)+(10*(p >> 4)));\
  } \
}

#define MSF_TO_LBA(MSF_T, LBA_T)\
{	\
  uint8 p;  \
   p = MSF_T.m; \
   LBA_T = ((0xf & p)+(10*(p >> 4)));\
   LBA_T *= 60; \
   p = MSF_T.s; \
   LBA_T += ((0xf & p)+(10*(p >> 4)));\
   LBA_T *= 75; \
   p = MSF_T.f; \
   LBA_T += ((0xf & p)+(10*(p >> 4)));\
}

PRAGMA_END_PACKED

int open_vcd(void);
void close_vcd(void);
int play_vcd_image(const char* fn);
void Start_Pbc();

#endif
