/* video_out_dc.c
 * Copyright (C) 2006-2008 Josh Sutherland <OneThirty8@aol.com>
 
 * This file is part of VC/DC, a free MPEG-2 video player.
 * See http://www.storm-studios.net/onethirty8/ for updates.
 *
 * VC/DC is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * VC/DC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */ 
/* Based on:
 * video_out_null.c
 * Copyright (C) 2000-2003 Michel Lespinasse <walken@zoy.org>
 * Copyright (C) 1999-2000 Aaron Holtzman <aholtzma@ess.engr.uvic.ca>
 *
 * This file is part of mpeg2dec, a free MPEG-2 video stream decoder.
 * See http://libmpeg2.sourceforge.net/ for updates.
 *
 * mpeg2dec is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * mpeg2dec is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
 
/*  TODO:  There are quite a few global variables in this file that should be made static. */
#include <kos.h>
#include "config.h"

#include <stdlib.h>
#include <inttypes.h>

#include "mpeg2.h"
#include "video_out.h"
#include "mpeg2convert.h"
#include "libvcd.h"
#include "gp.h"
#include "vcdbuttons.h"
/*
 * This is the kind of code one can get lost in, so I'll try to give a good synopsis of what we are doing.
 * I am also going to try to be as verbose as possible in my comments, because some of this code
 * is likely to look like gibberish to a great many people (at least those unfamiliar with the PVR
 * API from KOS, and those who haven't looked inside KOS to harvest code to tweak for specific purposes).
 * It may seem too verbose for experienced programmers, but I know how frustrating it can be trying to
 * traverse code that is doing stuff I don't understand, and well-commented code has helped me learn a
 * great deal.
 *
 * Some of this will look familiar to those who have used La Cible's lvfdc, as
 * a lot of what I'm doing here is following the same idea, and I have referred
 * to his code quite a bit.  Opening and closing the driver is probably about the same,
 * and I may have even have some of La Cible's code in those routines--I know the "vid_set"
 * variable name came from looking at his code.
 *
 * But, we must do things a bit differently than lvfdc because we want to use the PVR for
 * scaling the video to the proper size to support various resolutions.  Whereas lvfdc
 * is more geared toward showing short videos as in-game cut-scenes and such (and it is therefore
 * acceptable to expect a developer to understand and work within rigid size limitations), we've
 * got to accept that expecting the end-user to understand such a limitation is unrealistic, realize
 * that we don't much like such a limitation either (I don't want to re-encode my VCDs),
 * and then come up with a solution.
 *
 * So, because we need to resize most videos and don't have the time to do so in software, we must
 * deal with a limitation of the PVR (the return on the investment is worth it).  Texture width and
 * height must be a power of two.  So, we allocate textures that are large enough to cope with any video
 * we're likely to throw at it plus some extra space to make the PVR happy (ie, the smallest power-of-two
 * textures that will display a DVD or SVCD-sized movie).
 * But now we don't have a straight one-shot copy via pvr_txr_load because we don't want to copy the swarf
 * (and so we don't bother telling libmpeg2 to give us some extra space--what's the use?), so the vo_txr_load
 * function does the chopping off for us--it uses store queues as pvr_txr_load does, but after each line it
 * skips over (1024 - g_width) pixels before writing to the next line of our texture.
 *
 * I believe I have BlackAura (from the DCEmulation message board) to thank for the suggestion to use a
 * modified version of the sq_cpy function from KallistiOS.  Patbier is also to be thanked for suggesting a separate
 * thread for displaying the decoded data.
 *
 * Also, please note that while there are several video drivers defined in the header file, only the first
 * one is at all likely to work--I think the rest do absolutely nothing, although if/when I decide to extract
 * a library out of this Frankenstein's-monster-like conglomeration of code from all over the place (my mgplib
 * alone has parts transplanted from LAME and the main mpg123 library), I may add in a driver for those who want a
 * RGB565 option in addition to the current uyvy driver.
 */

  /* take_notes() and read_notes() are just how I see debugging information without a coder's cable. */
extern void read_notes();
extern void take_notes(char * noteid, char * line, ...);

extern int remote_mode;
extern int pbc_get_button();
extern RemoteControl_t * remotecontrol;

static int vid_set = 0;  /* Just an easy way for all the functions that might be called to know if we're initialized. 
			1 = set, 0 = not set.*/

static int got_new_video;  /* When we have new video to display, this will be set to 1. Otherwise, it should be 0. */

static void *incoming_video;  /* A pointer to the incoming video.  We should get it in a
				display-ready format (ie, uyvy).  When dc_draw_frame
				is called, it will point us to wherever the new picture is. */
				
static kthread_t *txr_thd;  /* We should have a function down below which keeps looping while vid_set is 1,
				and if got_new_video is 1, it'll tell vo_txr_load to get busy. */
				
static mutex_t video_mut = MUTEX_INITIALIZER; /* We're running the decoder and the video display in separate threads
					so we need to keep the dc_draw_frame function from moving the pointer
					to incoming_video before vo_draw_frame is done uploading to the PVR. */

static pvr_ptr_t vid_tex[2]; /* Pointers to our destination video textures.
			Double-buffered to avoid tearing or other mischief. */

static int tx_cur;  /* Will always be 1 or 0, to be used as:
                   *blah blah* vid_tex[tx_cur]; or *blah blah* hdr[tx_cur];
		    txr_cur = 1-txr_cur; */

static int tx_allocated = 0;
static int g_width, g_height;
static float g_u, g_v; /* Texture offsets for right/bottom edges of our display textures. */
static float g_top, g_bottom; /* Screen offsets for top/bottom edges of our video.  Will be 1 and 480 for full-screen,
				and 60 and 420 for widescreen. */

static pvr_poly_cxt_t cxt[2];  /* Polygon context and polygon header to tell the PVR what to do with what we give it. */
static pvr_poly_hdr_t hdr[2];  /* These will be compiled upon initialization of the video output driver, so we just need
			   to pass the correct polygon header when it's time to draw a frame. */

int warn_invalid_frames = 0; /* Our VCD/MPEG video playback functions will display this if an invalid button has been pressed. */
psd_area_t *highlight; /* In Extended PSD Selection Lists, we'll need to highlight the active area. */
#define PSD_XUNIT (640.0f / 256.0f)
#define PSD_YUNIT (480.0f / 256.0f)

/* vo_txr_load() Loads raw texture data from an SH-4 buffer into PVR RAM */

/* This is a hacked up version of the sq_cpy function from kos--
   see kernel/arch/dreamcast/hardware/sq.c */
/*    kernel/arch/dreamcast/hardware/sq.c
   (c)2001 Andrew Kieschnick */
   
static void vo_txr_load(void *src) {
/*got_new_video = 0;*/ /* It should be safe to say this while our mutex is locked--dc_draw_frame can't
			set it back to 1 until we are done copying our data. */

    size_t count = g_width * 2;
    count += (count %4);

    for(int i=0;i<g_height;i++)
        sq_cpy(vid_tex[tx_cur]+(i*2048), src+(i*count), count);

    sq_wait();
}

extern void vidclock_handle_frame();

static void draw_texture(int which)
{	    
/* Display the picture--ripped from the kos png example. */
//mutex_lock(&video_mut);
pvr_vertex_t vert;

//    vidclock_handle_frame();

    pvr_wait_ready();
    pvr_scene_begin();
       
    pvr_list_begin(PVR_LIST_OP_POLY);

    pvr_prim(&hdr[which], sizeof(hdr[which]));

    vert.argb = PVR_PACK_COLOR(1.0f, 1.0f, 1.0f, 1.0f);    
    vert.oargb = 0;
    vert.flags = PVR_CMD_VERTEX;
    
    vert.x = 1;
    vert.y = g_top;//1;
    vert.z = 1;
    vert.u = 0.0;
    vert.v = 0.0;
    pvr_prim(&vert, sizeof(vert));
    
    vert.x = 640;
    vert.y = g_top;//1;
    vert.z = 1;
    vert.u = g_u;
    vert.v = 0.0;
    pvr_prim(&vert, sizeof(vert));
    
    vert.x = 1;
    vert.y = g_bottom;//480;
    vert.z = 1;
    vert.u = 0.0;
    vert.v = g_v;
    pvr_prim(&vert, sizeof(vert));
    
    vert.x = 640;
    vert.y = g_bottom;//480;
    vert.z = 1;
    vert.u = g_u;
    vert.v = g_v;
    vert.flags = PVR_CMD_VERTEX_EOL;
    pvr_prim(&vert, sizeof(vert));
    
    pvr_list_finish();


    pvr_list_begin(PVR_LIST_TR_POLY);
#ifdef DEBUGFPS    
    read_notes();
#endif
    if (remote_mode >= 1)
    {
     int button_number;
     button_number = pbc_get_button();
      draw_poly_box(565.0f, 55.0f, 615.0f, 85.0f, 99.0f,
		0.80f, 0.0f, 0.0f, 0.0f,
		0.60f,  0.0f, 0.0f, 0.0f);
      if (button_number < 10)
      {		
      draw_poly_strf(580.0f, 60.0f, 100.0f, 1.0f, 1.0f, 1.0f, 0.0f,"0%d", button_number );
      }else{
      draw_poly_strf(580.0f, 60.0f, 100.0f, 1.0f, 1.0f, 1.0f, 0.0f,"%d", button_number );
      }			
    }
    if (remote_mode == 2)
    {
      draw_poly_mouse( 1+(remotecontrol->x * PSD_XUNIT), 1+(remotecontrol->y * PSD_YUNIT), 0.77f);
      if (!(highlight == NULL)) 
      {
       float tempx1, tempy1, tempx2, tempy2;
            
       tempx1 = highlight->x1 * PSD_XUNIT;
       tempx2 = highlight->x2 * PSD_XUNIT;
       tempy1 = highlight->y1 * PSD_YUNIT;
       tempy2 = highlight->y2 * PSD_YUNIT;
       draw_poly_box(tempx1, tempy1, tempx2, tempy2, 99.0f,
		0.40f, 0.6f, 0.7f, 1.0f,
		0.10f,  1.0f, 0.7f, 0.6f);
      }
    }
    
    if (warn_invalid_frames)
    {
       draw_poly_box(565.0f, 55.0f, 595.0f, 85.0f, 99.0f,
		1.0f, 1.0f, 1.0f, 1.0f,
		1.0f, 1.0f, 1.0f, 1.0f);
       draw_poly_strf(570.0f, 60.0f, 100.0f, 1.0f, 1.0f, 0.0f, 0.0f,"X");
       warn_invalid_frames--; 
    }    
    pvr_list_finish();

    pvr_scene_finish();
    /* tx_cur = 1-tx_cur; */    
//   mutex_unlock(&video_mut);
}

static int redraw_flag = 0;
static void* video_loop(void* yarr)  /* This loop will start when we open the video driver
					and keep looping until we close the driver. */
{
    while (vid_set == 1)
    {
        if (got_new_video > 0)
        {
            mutex_lock(&video_mut);
            if (redraw_flag)
                goto no_handle_frame;
            vidclock_handle_frame();
            no_handle_frame:
            redraw_flag = 0;
            draw_texture(got_new_video-1); /* Draw our texture. */
            got_new_video = 0;
            mutex_unlock(&video_mut);
        }
        thd_pass(); /* We're done here for a bit. */
    }
    return NULL;
}

/* IMPORTANT!  DON'T FORGET TO PUT CHECKS IN HERE!

We need to check for the following:
- Have we ever initialized the driver and allocated buffers? We should have a way to pass a dummy buffer.
- Is the video thread running?  If not, then LET'S NOT START IT!
 */
uint8_t * last_buf = NULL;
void redraw_frame() /* This is really a hack, but oh well. */
{
 if (last_buf == NULL) return; 
#if 0
 if (!vid_set)  /* This SHOULD NOT happen, but apparently it does sometimes. */
 {		
   vid_set = 1;
   got_new_video = 0;
   txr_thd = thd_create(0, video_loop, NULL);    /* Start our display thread. */
 }  
#else 
  if (!vid_set)
  {
    vo_txr_load(last_buf);
    draw_texture(tx_cur);
    tx_cur = 1-tx_cur;
    return;
  }
#endif  
retry_redraw:   
  if (got_new_video > 0)
  {
    goto retry_redraw;
  } 
    mutex_lock (&video_mut); /* Request a turn at exclusive access to
    				the incoming_video pointer and got_new_video. */
    
    vo_txr_load(last_buf);
    redraw_flag = 1;
    got_new_video = tx_cur+1;      /* Alert txr_thd that we have some new data for it. */
    
    mutex_unlock (&video_mut); /* Let txr_thd have access to incoming_video now that it
    				knows where to find it.*/
    tx_cur = 1-tx_cur;
}

static void dc_draw_frame (vo_instance_t * instance,
			     uint8_t * const * buf, void * id)
{ 

 if (!vid_set)  /* This SHOULD NOT happen, but apparently it does sometimes. */
 {		
   vid_set = 1;
   got_new_video = 0;
   txr_thd = thd_create(0, video_loop, NULL);    /* Start our display thread. */
 }  
retry:   
  if (got_new_video > 0)
  {
    goto retry;
  } 
    mutex_lock (&video_mut); /* Request a turn at exclusive access to
    				the incoming_video pointer and got_new_video. */
    
    vo_txr_load(buf[0]);
    last_buf = buf[0]; /* Point at the new data so we know where it is if we need to redraw. */
    got_new_video = tx_cur+1;      /* Alert txr_thd that we have some new data for it. */
    
    mutex_unlock (&video_mut); /* Let txr_thd have access to incoming_video now that it
    				knows where to find it.*/
    tx_cur = 1-tx_cur;         
}

static void dc_close (vo_instance_t * _instance)
{

    vid_set = 0; /* This will cause our looping thread to exit. */
	
    thd_join(txr_thd, NULL);  /* Now give it a chance to exit if it's still doing something. */
    
    if (tx_allocated ==1)
    {
      pvr_mem_free (vid_tex[0]);
      pvr_mem_free (vid_tex[1]);
      tx_allocated = 0;
    }  
    	    
    free ((void *)_instance);
    
}

static vo_instance_t * internal_open (int setup (vo_instance_t *, unsigned int,
						 unsigned int, unsigned int,
						 unsigned int,
						 vo_setup_result_t *),
				      void draw (vo_instance_t *,
						 uint8_t * const *, void *))
{
    vo_instance_t * instance;

    instance = (vo_instance_t *) malloc (sizeof (vo_instance_t));
    if (instance == NULL)
	return NULL;

    instance->setup = setup;
    instance->setup_fbuf = NULL;
    instance->set_fbuf = NULL;
    instance->start_fbuf = NULL;
    instance->draw = draw;
    instance->discard = NULL;
    instance->close = dc_close;//(void (*) (vo_instance_t *)) free;

    return instance;
}

static int dc_setup (vo_instance_t * instance, unsigned int width,
		       unsigned int height, unsigned int chroma_width,
		       unsigned int chroma_height, vo_setup_result_t * result)
{
    result->convert = mpeg2convert_uyvy; /* Libmpeg2 will use this function to convert the YUV planes into our
    					desired display format */
    
    /* If we're not set up or are given new values for width and height, get our values for our texture offsets. */

      g_u = (float)width / 1024.0f;
      g_v = (float)height / 1024.0f;
    
    g_width = width; g_height = height;  /* Just copy these local variables to our globals. */
    
    if (vid_set)  /* This shouldn't happen really, but in theory it could.  So, we need to wait for the running
    		     video thread to exit before we try to start it again. */
    {
      vid_set = 0;  /* This will cause the thread to exit. */
      thd_join(txr_thd, NULL);  /* Wait for it to exit. */
      vid_set = 1; /* Set this back.  Otherwise we'll allocate stuff that's already allocated! */
    }
    
    if (!vid_set)  /* If we're not initialized, we need to make sure we have a 640*480 screen, init the 3d hardware,
    			and allocate our destination textures and compile our polygon headers. */
    {
     
      /* If you're using this code in your own program, you may wish to uncomment these next two lines.  VC/DC already has us
      set up with a 640x480 16-bit video mode and the PVR is initialized when the menu loads up. Remember that you will have
      to reset any video mode you had previously set up!  Also, we require a ton of video memory here.  You will probably need
      to unload all of your textures and reload them when you're done playing videos, or alter this code so that your program
      uses smaller texture sizes. */
      
     // vid_set_mode (DM_640x480, PM_RGB565); 
     // pvr_init_defaults();
      
      vid_tex[0] = pvr_mem_malloc(1024*1024*2);
      vid_tex[1] = pvr_mem_malloc(1024*1024*2);
      
      pvr_poly_cxt_txr(&cxt[0], PVR_LIST_OP_POLY, PVR_TXRFMT_YUV422|PVR_TXRFMT_NONTWIDDLED, 1024, 1024, vid_tex[0], PVR_FILTER_BILINEAR);
      pvr_poly_compile(&hdr[0], &cxt[0]);
      
      pvr_poly_cxt_txr(&cxt[1], PVR_LIST_OP_POLY, PVR_TXRFMT_YUV422|PVR_TXRFMT_NONTWIDDLED, 1024, 1024, vid_tex[1], PVR_FILTER_BILINEAR);
      pvr_poly_compile(&hdr[1], &cxt[1]);
      tx_allocated = 1;
      
      vid_set = 1;  /* Here, we just need these variables set to the correct 'starting fresh' values. */
      tx_cur = 0;
      got_new_video = 0;
      
    }
    txr_thd = thd_create(0, video_loop, NULL);    /* Start our display thread. */
    return 0;
}

/* This is the function called by our main program to initialize our video driver. */
vo_instance_t * vo_dc_open (void)
{
    g_top = 1; g_bottom = 480;
    return internal_open (dc_setup, dc_draw_frame);
}

vo_instance_t * vo_dcws_open (void)
{
    g_top = 60; g_bottom = 420;
    return internal_open (dc_setup, dc_draw_frame);
}
/*
 *   There are some drivers listed below that will actually think they have the ability to get us set up--they don't.
 *   If you need a driver for a specific display format (eg, RGB565), fix it to operate like above, but with the correct
 *   mpeg2_convert_## function and using the correct flags when compiling your polygon headers.  I strongly suggest, however,
 *   that you stick with uyvy (YUV422) as the conversion to this format is a matter of packing the three Y/U/V planes into one
 *   buffer, which is simpler and faster than yuv->rgb conversion.
 *   
 */

#if 0   
vo_instance_t * vo_dcskip_open (void)
{
    return internal_open (dc_setup, NULL);
}

static void dcslice_start (void * id, const mpeg2_fbuf_t * fbuf,
			     const mpeg2_picture_t * picture,
			     const mpeg2_gop_t * gop)
{
}

static void dcslice_copy (void * id, uint8_t * const * src,
			    unsigned int v_offset)
{
}

static int dcslice_convert (int stage, void * id,
			      const mpeg2_sequence_t * seq,
			      int stride, uint32_t accel, void * arg,
			      mpeg2_convert_init_t * result)
{
    result->id_size = 0;
    result->buf_size[0] = result->buf_size[1] = result->buf_size[2] = 0;
    result->start = dcslice_start;
    result->copy = dcslice_copy;
    return 0;
}

static int dcslice_setup (vo_instance_t * instance, unsigned int width,
			    unsigned int height, unsigned int chroma_width,
			    unsigned int chroma_height,
			    vo_setup_result_t * result)
{
    result->convert = dcslice_convert;
    return 0;
}

vo_instance_t * vo_dcslice_open (void)
{
    return internal_open (dcslice_setup, dc_draw_frame);
}

static int dcrgb16_setup (vo_instance_t * instance, unsigned int width,
			    unsigned int height, unsigned int chroma_width,
			    unsigned int chroma_height,
			    vo_setup_result_t * result)
{
    result->convert = mpeg2convert_rgb16;
    if (!vid_set)
    {
      vid_set_mode(DM_320x240, PM_RGB565);
      vid_set = 1;
    }
    return 0;
}

static int dcrgb32_setup (vo_instance_t * instance, unsigned int width,
			    unsigned int height, unsigned int chroma_width,
			    unsigned int chroma_height,
			    vo_setup_result_t * result)
{
    result->convert = mpeg2convert_rgb32;
    if (!vid_set)
    {
      vid_set_mode(DM_320x240, PM_RGB888);
      vid_set = 1;
    }
    return 0;
}

vo_instance_t * vo_dcrgb16_open (void)
{
    return internal_open (dcrgb16_setup, dc_draw_frame);
}

vo_instance_t * vo_dcrgb32_open (void)
{
    return internal_open (dcrgb32_setup, dc_draw_frame);
}
#endif
