/*
 * mpeg2play.c
 * Copyright (C) 2006 - 2009 Josh Sutherland <OneThirty8@aol.com>
 * This file is part of VC/DC, a free MPEG-2 video player.
 *
 * VC/DC is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * VC/DC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
/*
 * based on mpeg2dec.c
 * Copyright (C) 2000-2003 Michel Lespinasse <walken@zoy.org>
 * Copyright (C) 1999-2000 Aaron Holtzman <aholtzma@ess.engr.uvic.ca>
 *
 * This file is part of mpeg2dec, a free MPEG-2 video stream decoder.
 * See http://libmpeg2.sourceforge.net/ for updates.
 *
 * mpeg2dec is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * mpeg2dec is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 *  
 *   A caveat for anyone looking at this code--look closely for preprocessor directives, particularly
 *   "#if 0" because I tend to use that a lot when trying different ways of doing things.  The v0.2 release
 *   will probably have some of that in it--I'm working on several things, including PBC for VideoCD, and migrating
 *   the vidclock_t struct and associated functions to a different source code module so that this file will be strictly
 *   demultiplexing functions, functions to interface with the video and audio decoders, and the three (or four, if we count
 *   VCD) that make use of those functions.  In the interim, there is a whole lot in this file that really belongs elsewhere.
 *   It WILL be cleaned up, and much of the messiness is actually due to my attempts at working toward a cleaner interface here
 *   (see the use of #if 0 in the decode_mpeg2 function--framerate calculation _should_not_be_ the job of that function, so it
 *   has been moved to viclock_handle_incoming and vidclock_handle_frame (which isn't called from within this module, but from our
 *   libvo driver when it wants to know if it's okay to display the last frame it has received).
 *   
 *   On the plus side, I've managed to accomplish increase in playback speed (mainly by using separate threads for display/sound
 *   output), a decrease in number of dropped frames due to a better method of calculating framerate at decode time, and smoother
 *   playback on streams that used to be very jerky by slowing us down at display time (which we get away with because that happens
 *   in a separate thread and control is then passed to the next thread waiting to do stuff).
 *  
 */

#include <kos.h>
#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
//#include <getopt.h>
#ifdef HAVE_IO_H
#include <fcntl.h>
#include <io.h>
#endif


#include "mpeg2.h"
#include "video_out.h"
//#include "gettimeofday.h"

#include "mpglib_config.h"
#include "../mpglib/mpg123.h"
#include "../mpglib/output.h"
#include "../mpglib/interface.h"
#include "../mpglib/decode_i386.h"
#include "gp.h"
//#include "id3.h"
#include "libvcd.h"
#include "mp3junk.h"
#include "mpegav.h"
#include "../vidclock/vidclock.h"
extern uint8_t * vcd_buffer;    
/* libmpeg2 video decoder stuff */
static int buffer_size = 32768/2;//4096; 
				/*A larger buffer should give smoother playback
				 on the DC, since we'll run out of stuff to play
				 less often.  I think.*/
static file_t in_file;
static int demux_track = 0;
static int demux_pid = 0;
static mpeg2dec_t * mpeg2dec;
static vo_open_t * output_open = NULL;
static vo_instance_t * output = NULL;
static int sigint = 0;
static int total_offset = 0;
//static int verbose = 0;

/* mpeg audio stuff */
#define BUFFER_SIZE 4096
static int mpeg_audio_activated;
struct mpstr_tag mp;
void *fo;
char mpa_out[65536*2];/*char mpa_out[8192];*/
char resample_out[65536*2];
struct audio_out *audio_out = &oss_out;
int apts, vpts, adts, vdts, cpts, cdts;

int seeking = 0;  /* We should make this a private variable in the vidclock struct with functions
			to set and get. */

/*functions start here */
/* void dump_state (FILE * f, mpeg2_state_t state, const mpeg2_info_t * info,
		 int offset, int verbose); */
		 
/* playback control stuff */
//int cur_chapter;

/* static RETSIGTYPE signal_handler (int sig)
{
    sigint = 1;
    signal (sig, SIG_DFL);
    return (RETSIGTYPE)0;
} */

int st_fl;
int fr_p;

#ifdef DEBUGFPS
extern void take_notes(char * noteid, char * fmt, ...);
extern void clear_note(char * noteid);
extern void clear_notes();
#endif

extern void fudge_audio(double fudge);

#if 0 /* This stuff is getting moved to a separate vidclock module. */
#define PTIME_CHUNK 1.0f/15.0f /* For VCD 2.0 support, item playtime units. */
struct vidclock_s{
  uint32_t frame_counter;
  uint32_t last_count;
  double fps;
  double cur_fps; /* We'll keep track of this for about 1/8 of a second to decide whether or
  			not to drop video frames. */
  double ptime_inc; /* Useful for supporting ptime in VCD PlayLists as well as a/v synch. */
  
  double ptime_total; /* How many ptime units have passed since starting playback. */
  mutex_t * display_mut;
  
  int incoming_reset, decoded_reset; /* Set these to 1 if we just started/skipped/whatever. */
  int skipped;  /* Useful if debugging--throw some code in somewhere to see how many frames you're skipping. */
  struct timeval recent_frames[6]; /* Here we keep track of the framerate over six frames.  If we skip a
  						frame, we need to update it anyway so the video output thread knows
						how to fake the 'real' framerate. */
  /* Framerate regulation stuff */
  uint32_t esb; /* 'elapsed should be __' in other words, how long should elapse between frames */ 
  
  int frame_period; /* Really only useful for debugging... */

  uint64_t last_frame_time; /* The time it was last time we got a frame. Useful for closer a/v synch. */
  int audio_rate;  /* The size, in samples, of a frame of audio. */
  int sample_counter;
  int audio_latency;
} vidclock;

static void fps_init(unsigned int frame_period)
{
 double temp;
  vidclock.frame_counter = vidclock.last_count = 0;
  
  /* Frame period is stored assuming a 27 MHz clock.
  Dividing 27 million by the frame period will give us the
  frame rate. */
  vidclock.fps = 27000000.0 / frame_period;
  vidclock.cur_fps = vidclock.fps;
  
 /* Calculate the ptime value for this sequence. */
   vidclock.ptime_total = 0.0f;
   vidclock.ptime_inc = (1.0f/vidclock.fps);
 
 /* Calculate the framerate regulation value. */
  temp = 100.0 / (vidclock.fps ); /* Do we want to round up? */
  vidclock.esb = (int)(temp);//*4;  /* Check every four frames or so. */
  
  if (vidclock.display_mut == NULL)
     mutex_create(vidclock.display_mut);
  
  vidclock.incoming_reset = vidclock.decoded_reset = 1;
  vidclock.audio_rate = 0;  /* If we just started, we have no clue what this is. */
  vidclock.sample_counter = 0;
  vidclock.last_frame_time = timer_ms_gettime64();
#ifdef DEBUGFPS
clear_notes();
#endif  
}

/* This is good enough for video-only sequences, but for streams with audio we can
   probably be even more accurate by looking at the number of samples that have played. */
int vidclock_get_ptime()
{
  double chunks;
  chunks = (vidclock.ptime_total * 15.0f);
  return (uint16)chunks;
}

void fps_reset()
{
  vidclock.cur_fps = vidclock.fps;
  vidclock.last_count = vidclock.frame_counter =
  vidclock.audio_rate = vidclock.sample_counter = 0;
  
  vidclock.incoming_reset = vidclock.decoded_reset = 1;
}

void vidclock_set_audio_rate(int freqs, int chans, int buf_size)
{
  vidclock.audio_rate = freqs;
  
  /* Account for latency in samples by dividing the buffer size by 2
     times the number of channels. */ 
  vidclock.audio_latency = buf_size/(chans*2);  
}

int vidclock_drift_allowance()
{
 if(vidclock.audio_rate && vidclock.fps)
  return(int)(vidclock.audio_rate/(vidclock.fps*2)); /* Basically, half a frame isn't too horribly out of synch. */
 else
  return 0;
}

/* We should only call this once per call to the audio decoder function. Also, rather than synch audio to video,
  we'll probably do better to synch to an external clock. We just need to figure a way to keep the clock from going
  wonky if the disc has spun down during a pause or something. */
static int vidclock_handle_audio(uint32_t samples)
{
  double vidsecs; /* How much time, in seconds, has passed since the last frame was shown/skipped. */
  uint64_t now;
  int vidsecs_to_samples;
  static int last_framenum = 0; /* Use this to check that we don't resample too often! */
   
   vidclock.sample_counter += samples;
   
   if (vidclock.frame_counter == 0)
   {
     last_framenum = 0;
     vidclock.sample_counter = 0;
     vidclock.last_frame_time = timer_ms_gettime64(); /* Not perfect, but it will drift back quickly. */
     return 0;
   }
   if ((vidclock.frame_counter - last_framenum) * vidclock.ptime_inc < 0.5) return 0; /* So we don't resample too often. */
   
   last_framenum = vidclock.frame_counter;
   
   now = timer_ms_gettime64();
   vidsecs = vidclock.ptime_total + (((double)(now - vidclock.last_frame_time))/1000.0f);
   vidsecs_to_samples = (int)(vidsecs * vidclock.audio_rate);
   //vidsecs_to_samples = (int)(vidclock.ptime_total * vidclock.audio_rate);
   
   /* A positive value means that the audio is behind, negative that it's
   ahead. */
   return vidsecs_to_samples - vidclock.sample_counter;   
}

static void vidclock_handle_incoming (const mpeg2_info_t * info)
{
    static uint32_t frame_counter = 0;
    static struct timeval tv_beg[6], tv_start;  /* Having 6 tv_beg timevalues will allow us to keep track of
    						the framerate for ~ 1/4 second (NTSC-Film).  This should be enough? */
    static int total_elapsed;
    static int last_count = 0;
    struct timeval tv_end;
    static double fps, tfps;
    int frames, elapsed;
    int i;

    if (verbose)
	return;

    gettimeofday (&tv_end, NULL);

    if (vidclock.incoming_reset==1){
        frame_counter = 0; vidclock.cur_fps = vidclock.fps; vidclock.skipped = 0;
	tv_start = tv_beg[0] = tv_beg[1] = tv_beg[2] = tv_beg[3] = tv_beg[4] = tv_beg[5] =
	tv_end;
	signal (SIGINT, signal_handler);
	vidclock.incoming_reset = 0;
    }
    

    elapsed = (tv_end.tv_sec - tv_beg[0].tv_sec) * 100 +
	(tv_end.tv_usec - tv_beg[0].tv_usec) / 10000;
    total_elapsed = (tv_end.tv_sec - tv_start.tv_sec) * 100 +
	(tv_end.tv_usec - tv_start.tv_usec) / 10000;


    frame_counter++;

    for ( i = 0; i < 5; i++)
    {
      tv_beg[i] = tv_beg[i+1];
    } 
    
    tv_beg[5] = tv_end;
   
    frames = ( frame_counter < 6) ? frame_counter : 6;

    fps = frames * 100.0 / elapsed;
    tfps = frame_counter * 100.0 / total_elapsed;
#ifdef DEBUGFPS  
    take_notes("decoded framerate display", "Decoded: t:%.2f/c:%.2f", tfps, fps);
#endif    		
    last_count = frame_counter;   
    

incoming_end: 
    /* skip B-frames if behind */ 
    if (
       ( ( (info->current_picture->flags &  PIC_MASK_CODING_TYPE) == PIC_FLAG_CODING_TYPE_B) &&
                                                                         ( (fps/vidclock.fps < 1.0) && (!(tfps/vidclock.fps > 1.0)) ) )
       || /* or skip all but every nth I-frame if we are fast-forwarding. */
       ( (seeking==1) && ( ! ( (info->current_picture->flags &  PIC_MASK_CODING_TYPE) == PIC_FLAG_CODING_TYPE_I) ) )
       )
    {
      mpeg2_skip(mpeg2dec, 1);

     vidclock.skipped ++;
            
    }else{
      mpeg2_skip(mpeg2dec, 0);
    }
    

#ifdef DEBUGFPS
     take_notes("fps skipcount","%d of %d skipped = %.2f", vidclock.skipped, frame_counter, ((double)vidclock.skipped/(double)frame_counter)) ;
#endif
   
}

/* I don't think this is really totally accurate--if we're skipping a frame and writing a new timeval to
   vidclock.recent_frames[5], that's overwriting the sixth frame that vidclock_handle_frame() is looking at...
   so I think we really have a span of seven frames somewhere in there, but we're saying six.  This may be why
   we drift in and out of synch by a few frames sometimes?  Or my Star Wars VCDs are encoded slightly out of synch--
   the video looks few frames ahead on my DVD player sometimes too, so I guess this works well enough for now, but in
   the future we should probably revisit this and decide if we need to figure a way to make this more accurate. */
   
/* Should we do this at all? We should be able to compensate well enough when we synch audio since we account for the
  time that has elapsed since the last frame displayed. If, instead of saying in the vidclock_handle_frame function
  vidclock.ptime_total += vidclock.ptime_inc, we say vidclock.ptime_total = vidclock.framecounter * vidclock.ptime_inc,
  I think we may fare better.  We should probably also only need two timeval structs, or even just a millisecond counter
  since this program isn't exactly portable anyway at this point.  */
   
void vidclock_skip_frame()
{

   mutex_lock(vidclock.display_mut);
   
   gettimeofday (&vidclock.recent_frames[5], NULL);
   
   vidclock.ptime_total += vidclock.ptime_inc;
   vidclock.last_frame_time = timer_ms_gettime64(); 
   mutex_unlock(vidclock.display_mut);
}

/* We should be able to make this a whole lot simpler and probably more accurate in the bargain.  Get
  rid of the timeval structs and just use last_count and vidclock.frame_counter, vidclock.last_frame_time
  and the current time reported by timer_ms_gettime64();. */
void vidclock_handle_frame ()
{
    static uint32_t frame_counter = 0;
    static struct timeval tv_beg, tv_start;
    static int total_elapsed;
    static int last_count = 0;
    struct timeval tv_end;
    static double fps, tfps;
    int frames, elapsed;
    int i;

    if (verbose)
	return;

    gettimeofday (&tv_end, NULL);

    vidclock.frame_counter--; /* A little trick! */ 
    /* I am no longer sure what the trick is--may be a leftover from something else I tried. Now,
    we do a check on something else, lock a mutex, and re-increment the framecounter. */
    
    if (vidclock.decoded_reset==1){
	tv_start = tv_beg = tv_end;
	vidclock.recent_frames[0] = vidclock.recent_frames[1] = vidclock.recent_frames[2] = 
	vidclock.recent_frames[3] = vidclock.recent_frames[4] = vidclock.recent_frames[5] = tv_end; 
	signal (SIGINT, signal_handler);
	vidclock.decoded_reset = 0;
    }
  mutex_lock(vidclock.display_mut);  
    
    vidclock.ptime_total += vidclock.ptime_inc;
    vidclock.frame_counter++;

vc_wait:
	
    elapsed = (tv_end.tv_sec - vidclock.recent_frames[0].tv_sec) * 100 +
	(tv_end.tv_usec - vidclock.recent_frames[0].tv_usec) / 10000;	
    total_elapsed = (tv_end.tv_sec - tv_start.tv_sec) * 100 +
	(tv_end.tv_usec - tv_start.tv_usec) / 10000;
    
    frames = (vidclock.frame_counter < 6) ? vidclock.frame_counter : 6;

    vidclock.cur_fps = frames * 100.0 / elapsed;
    tfps = vidclock.frame_counter * 100.0 / total_elapsed;
    
     if ( (vidclock.cur_fps/vidclock.fps ) > 1.0)
    {
      gettimeofday(&tv_end, NULL);
      thd_pass();
      goto vc_wait;
    } 
    
    for ( i = 0; i < 5; i++)
    {
      vidclock.recent_frames[i] = vidclock.recent_frames[i+1];
    }
    
    vidclock.recent_frames[5] = tv_beg = tv_end;
    mutex_unlock(vidclock.display_mut);
#ifdef DEBUGFPS   
    take_notes("simulated framerate display", "This frame: t:%.2f/c:%.2f /Should be:%.2f", tfps, vidclock.cur_fps, vidclock.fps);    
#endif 		
    last_count = vidclock.frame_counter;
    vidclock.last_frame_time = timer_ms_gettime64();
}        
#endif /* if 0 */

static void print_fps (int final)
{
#ifdef DEBUGFPS
    static uint32_t frame_counter = 0;
    static struct timeval tv_beg, tv_start;
    static int total_elapsed;
    static int last_count = 0;
    struct timeval tv_end;
    static double fps, tfps;
    int frames, elapsed;
    

    if (verbose)
	return;

    gettimeofday (&tv_end, NULL);

    if (!frame_counter) {
	tv_start = tv_beg = tv_end;
	signal (SIGINT, signal_handler);
    }

    elapsed = (tv_end.tv_sec - tv_beg.tv_sec) * 100 +
	(tv_end.tv_usec - tv_beg.tv_usec) / 10000;
    total_elapsed = (tv_end.tv_sec - tv_start.tv_sec) * 100 +
	(tv_end.tv_usec - tv_start.tv_usec) / 10000;

    if (final) {
	if (total_elapsed)
	    tfps = frame_counter * 100.0 / total_elapsed;
	else
	    tfps = 0;

	fprintf (stderr,"\n%d frames decoded in %.2f seconds (%.2f fps)\n",
		 frame_counter, total_elapsed / 100.0, tfps);
		 
	frame_counter = 0;
	return;
    }

    frame_counter++;

    if (elapsed < 50)	/* only display every 0.50 seconds */
    {
	return;
    }
    tv_beg = tv_end;
    frames = frame_counter - last_count;

    fps = frames * 100.0 / elapsed;
    tfps = frame_counter * 100.0 / total_elapsed;
    
    take_notes("real framerate display", "Displayed: t:%.2f/c:%.2f", tfps, fps);
    		
    last_count = frame_counter;
#endif
}

#if 0
static void * malloc_hook (unsigned size, mpeg2_alloc_t reason)
{
    void * buf;

    /*
     * Invalid streams can refer to fbufs that have not been
     * initialized yet. For example the stream could start with a
     * picture type other than I. Or it could have a B picture before
     * it gets two reference frames. Or, some slices could be missing.
     *
     * Consequently, the output depends on the content 2 output
     * buffers have when the sequence begins. In release builds, this
     * does not matter (garbage in, garbage out), but in test code, we
     * always zero all our output buffers to:
     * - make our test produce deterministic outputs
     * - hint checkergcc that it is fine to read from all our output
     *   buffers at any time
     */
    if ((int)reason < 0) {
        return NULL;
    }
    buf = mpeg2_malloc (size, (mpeg2_alloc_t)-1);
    if (buf && (reason == MPEG2_ALLOC_YUV || reason == MPEG2_ALLOC_CONVERTED))
        memset (buf, 0, size);
    return buf;
}
#endif
/* We should also figure a decent way to handle different aspect ratios and aspect ratio
   changes.  This may mean modifying libvo. */

static void shut_video_out()
{
 if (output->close)
	output->close (output);
  output = NULL;
}   

/* Based on mpeg2_guess_aspect (const mpeg2_sequence_t * sequence, unsigned int * pixel_width,
			unsigned int * pixel_height) */
static void set_video_aspect(const mpeg2_sequence_t * sequence)
{
 int DAR_16_9;
  vo_driver_t const * drivers;
  
  if (!(output==NULL)) shut_video_out();
   
  drivers = vo_drivers();
  
  if (! (sequence->flags & SEQ_FLAG_MPEG2)) {
	DAR_16_9 = (sequence->pixel_height == 27 ||
		    sequence->pixel_height == 45);
    } else {
	DAR_16_9 = (3 * sequence->picture_width * sequence->pixel_width >
		    4 * sequence->picture_height * sequence->pixel_height);
    }
  if (DAR_16_9)				/* 16:9 aspect ratio. */
  {           
   output_open = drivers[1].open;
  }else{         			/* Assume a 4:3 aspect ratio (should be good enough). */
   output_open = drivers[0].open;
  }
   
  output = output_open ();
}    
       
static void decode_mpeg2 (uint8_t * current, uint8_t * end)
{
    const mpeg2_info_t * info;
    mpeg2_state_t state;
    vo_setup_result_t setup_result;

    mpeg2_buffer (mpeg2dec, current, end);
    total_offset += end - current;

    info = mpeg2_info (mpeg2dec);
	     
    while (1) {
	    
	state = mpeg2_parse (mpeg2dec);
	/* if (verbose)
	    dump_state (stderr, state, info,
			total_offset - mpeg2_getpos (mpeg2dec), verbose); */
	switch (state) {
	case STATE_BUFFER:
	    return;
	case STATE_SEQUENCE:
	    set_video_aspect(info->sequence);
	    /* might set nb fbuf, convert format, stride */
	    /* might set fbufs */
	    if (output->setup (output, info->sequence->width,
			       info->sequence->height,
			       info->sequence->chroma_width,
			       info->sequence->chroma_height, &setup_result)) {
		fprintf (stderr, "display setup failed\n");
		exit (1);
	    }
	    if (setup_result.convert &&
		mpeg2_convert (mpeg2dec, setup_result.convert, NULL)) {
		fprintf (stderr, "color conversion setup failed\n");
		exit (1);
	    }
	    if (output->set_fbuf) {
		uint8_t * buf[3];
		void * id;

		mpeg2_custom_fbuf (mpeg2dec, 1);
		output->set_fbuf (output, buf, &id);
		mpeg2_set_buf (mpeg2dec, buf, id);
		output->set_fbuf (output, buf, &id);
		mpeg2_set_buf (mpeg2dec, buf, id);
	    } else if (output->setup_fbuf) {
		uint8_t * buf[3];
		void * id;

		output->setup_fbuf (output, buf, &id);
		mpeg2_set_buf (mpeg2dec, buf, id);
		output->setup_fbuf (output, buf, &id);
		mpeg2_set_buf (mpeg2dec, buf, id);
		output->setup_fbuf (output, buf, &id);
		mpeg2_set_buf (mpeg2dec, buf, id);		
	    }
	    mpeg2_skip (mpeg2dec, (output->draw == NULL));
	    	    
	    vidclock_init(info->sequence->frame_period);

	    break;
	case STATE_PICTURE:
	    /* might skip */
            vidclock_handle_incoming (info, mpeg2dec);
	    /* might set fbuf */
	    if (output->set_fbuf) {
		uint8_t * buf[3];
		void * id;

		output->set_fbuf (output, buf, &id);
		mpeg2_set_buf (mpeg2dec, buf, id);
	    }
	    if (output->start_fbuf)
		output->start_fbuf (output, info->current_fbuf->buf,
				    info->current_fbuf->id);
	    break;
	case STATE_SLICE:
	case STATE_END:
	case STATE_INVALID_END:
  	    /* might free frame buffer */
	    /* vidclock.frame_counter++; */
	    		
	    if ( (info->display_fbuf) && (!(info->display_picture->flags & PIC_FLAG_SKIP)) ) {
		 if (output->draw) 
		    output->draw (output, info->display_fbuf->buf,
				  info->display_fbuf->id);
		print_fps (0);

	    } else {
	      vidclock_skip_frame();
	    }
	    if (output->discard && info->discard_fbuf)
		output->discard (output, info->discard_fbuf->buf,
				 info->discard_fbuf->id);
	  
	    break;	    
	default:
	    break;	    
	    
	}
	   
    }
    
}
	             	    	    


void mpa_reset() 
{
    if (mpeg_audio_activated==1)
    {
      ExitMP3(&mp);
      audio_out->close(audio_out);
      mpeg_audio_activated = 0;
    }   

    print_fps(1);
}

/* This stuff allows us to use some functions in my Frankenstein-like mpglib that I borrowed from the
  main mpg123 library--instead of decoding the mpeg audio at its full sampling rate, we can decode every
  second or fourth sample instead.  I don't know how much of an increase in 'percieved video quality'
  we get out of this, but it does work on videos that we need to skip frames on, and we won't skip quite
  as many frames if we use half-quality audio.  We do retain the stereo image (on files with stereo audio),
  so things like Pro-Logic Surround that are present in the encoded stream should still work, but the
  overall sound quality will suffer some.  Using synth_2to1 is acually not bad (but noticably poorer than
  synth_1to1), but synth_4to1 is awful.  To my ears, anyway.  Your mileage may vary.  I may try to hack in a
  bass-boost/treble-reduce to help make things sound a tad better (I noticed an odd high-frequency ringing
  effect when testing with synth_4to1 so the treble cut would help make that less noticable), but then we're
  spending time doing the EQ which kinda defeats the whole 'spend less time on audio decoding' thing. */
  
int (* synth_mono)(PMPSTR mp, real *bandPtr,unsigned char *out,int *pnt);
int (* synth_)(PMPSTR mp, real *bandPtr,int channel,unsigned char *out,int *pnt);
int audio_quality = 0;

void set_audio_quality(int quality)
{
  switch (quality)
  {
    case 1:
      audio_quality = 1;
      synth_mono = synth_1to1_mono;
      synth_ = synth_1to1;
    break;
    case 2:
      audio_quality = 2;
      synth_mono = synth_2to1_mono;
      synth_ = synth_2to1;
    break;
    /* Assume that if somebody calls this with '3' or '4' for quality, they mean 'worse than 2.' */
    case 3: 
    case 4:
      audio_quality = 4;
      synth_mono = synth_4to1_mono;
      synth_ = synth_4to1;
    break; 
    default:
      audio_quality = 0; /* This will cause the mpa_decode_data function to default back to 1, but whatever. */
    break;
  }
}
/* mpa_decode_data:
    start is the beginning of the current hunk of data.
    end is the end of the current hunk of data.
    doneframe is a flag to let us know that it's OK to send
    all of the decoded data from this frame to the sound
    hardware.*/
#if 0
void mpa_decode_data (uint8_t * start, uint8_t * end, int doneframe)
{
 int len, ret, size;

   len = end - start;
    if(len <= 0)
        return;
  
  if (seeking==1) return;
  
  if (mpeg_audio_activated==0){
   if (audio_quality == 0){
     synth_ = synth_1to1;
     synth_mono = synth_1to1_mono;
     audio_quality = 1;
   }
   InitMP3(&mp);
  }
 
    ret = decodeMP3_clipchoice(&mp,start,len,mpa_out,&size,synth_mono, synth_);
 	st_fl = mp.fr.stereo;
   if (mpeg_audio_activated==0){
   fprintf(stderr, "frequency = %d and stereo = %d\n", freqs[mp.fr.sampling_frequency],mp.fr.stereo);
    fo = audio_out->open(NULL, freqs[mp.fr.sampling_frequency]/audio_quality, mp.fr.stereo, AFMT_S16_LE, 0, 8192);
    mpeg_audio_activated=1;
  }
    while(ret==MP3_OK)
    {
          audio_out->write(fo,mpa_out,size);
	  ret = decodeMP3_clipchoice(&mp,NULL,0,mpa_out,&size,synth_mono, synth_);	
    } 

    if (doneframe==1)  /* If we're ready to upload to the sound hardware...*/
        audio_out->write(fo,mpa_out, -1);  /*...we'll give an impossible
						number for size. */
						
}
#else

#define difference(a, b) ((a > b) ? a-b : b-a)
void (* resample_func)(int16_t* input_samples, uint32_t in_samples, 
				  int16_t* output_samples, uint32_t out_samples);
				  				
void mpa_decode_data (uint8_t * start, uint8_t * end, int doneframe)
{
 int len, ret, size;
 char *ptr;
 uint32 samples;
// static int audio_threshold = 0;
 static int total_size = 0;
   len = end - start;
    if(len <= 0)
        return;
  
  if (seeking==1) return;
  
  if (mpeg_audio_activated==0){
   if (audio_quality == 0){
     synth_ = synth_1to1;
     synth_mono = synth_1to1_mono;
     audio_quality = 1; 
   }
   total_size = 0;
   InitMP3(&mp);
  }

    ptr = mpa_out + total_size;
    ret = decodeMP3_clipchoice(&mp,start,len,mpa_out,&size,synth_mono, synth_);
 	st_fl = mp.fr.stereo;
   if (mpeg_audio_activated==0){
   fprintf(stderr, "frequency = %d and stereo = %d\n", freqs[mp.fr.sampling_frequency],mp.fr.stereo);
    fo = audio_out->open(NULL, freqs[mp.fr.sampling_frequency]/audio_quality, mp.fr.stereo, AFMT_S16_LE, 0, 8192);
    //fo = audio_out->open(NULL, freqs[mp.fr.sampling_frequency]/audio_quality, mp.fr.stereo, AFMT_S16_LE, 0, freqs[mp.fr.sampling_frequency]/audio_quality * mp.fr.stereo);
    vidclock_set_audio_rate(freqs[mp.fr.sampling_frequency]/audio_quality, mp.fr.stereo, 8192);
    mpeg_audio_activated=1;
    //resample_func = (mp.fr.stereo == 1) ? _x_audio_out_resample_mono : _x_audio_out_resample_stereo;
    resample_func = (mp.fr.stereo == 1) ? stretch_mono : stretch_stereo;
    //audio_threshold = 0;
  }
  
 // if(!audio_threshold) audio_threshold = vidclock_drift_allowance();
  
    while(ret==MP3_OK)
    {
	   ptr +=size; total_size += size;
	  ret = decodeMP3_clipchoice(&mp,NULL,0,ptr,&size,synth_mono, synth_);	
    }
  if (doneframe==1)
  {
   int sample_resync, new_size, min_size, max_size;
   
    	  samples = total_size/(2 * mp.fr.stereo);
	  sample_resync = vidclock_handle_audio(samples);
	  
	  min_size = samples * 90 / 100;
	  max_size = samples * 110 / 100;
	  if (samples + sample_resync < min_size) sample_resync = min_size - samples;
	  if (samples + sample_resync > max_size) sample_resync = max_size - samples;
	    new_size = total_size + (sample_resync*(2*mp.fr.stereo));
	    
	    /* if (new_size > total_size + 1000/audio_quality) new_size = total_size + 1000/audio_quality; */
	    /* if (new_size > total_size + total_size/4) new_size = total_size + total_size/4;
	    if (new_size < 16) new_size = 16; */ /* Make sure it's worth bothering to resample... */ 
	    
	    /* if (difference(new_size, total_size) < audio_threshold) new_size = total_size;
	    if (difference(new_size, total_size) > audio_threshold*2)
	    			new_size = (new_size > total_size) ? total_size + audio_threshold*2 : total_size - audio_threshold*2; 
		 */		
	    if (new_size < 16) new_size = 16; 
	    if (!(new_size==total_size))
	    {
	   	resample_func( (int16_t*) mpa_out, total_size/(2 * mp.fr.stereo), 
				  (int16_t*) resample_out, new_size/(2 * mp.fr.stereo));
				  
		audio_out->write(fo,resample_out,new_size);
		/* Below is alternative code that just lops off samples or pads out the buffer. */
		/* if (new_size > total_size)
		{
		  int i; uint16_t *samp_ptr;
		  samp_ptr = (uint16_t *)mpa_out;
		  i = total_size/2;
		  do
		  {
		    if (mp.fr.stereo > 1)
		    {
		      samp_ptr[i] = samp_ptr[total_size/2 - 2];
		      i++;
		    }
		    samp_ptr[i] = samp_ptr[total_size/2 - 1];
		    i++;
		  } while (i < new_size/2);
		}
		audio_out->write(fo,mpa_out, new_size); */				  
	    }else{
		audio_out->write(fo,mpa_out, new_size);
	    }
	total_size = 0;
   }
				
}
#endif

#define DEMUX_PAYLOAD_START 1
/*static */int demux (uint8_t * buf, uint8_t * end, int flags)
{
    static int mpeg1_skip_table[16] = {
	0, 0, 4, 9, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
    };

    /*
     * the demuxer keeps some state between calls:
     * if "state" = DEMUX_HEADER, then "head_buf" contains the first
     *     "bytes" bytes from some header.
     * if "state" == DEMUX_DATA, then we need to copy "bytes" bytes
     *     of ES data before the next header.
     * if "state" == DEMUX_SKIP, then we need to skip "bytes" bytes
     *     of data before the next header.
     *
     * NEEDBYTES makes sure we have the requested number of bytes for a
     * header. If we dont, it copies what we have into head_buf and returns,
     * so that when we come back with more data we finish decoding this header.
     *
     * DONEBYTES updates "buf" to point after the header we just parsed.
     */

#define DEMUX_HEADER 0
#define DEMUX_DATA 1
#define DEMUX_SKIP 2
#define DEMUX_MPA_DATA 3
    static int state = DEMUX_SKIP;
    static int state_bytes = 0;
    static uint8_t head_buf[264];

    uint8_t * header;
    int bytes;
    int len;

#define NEEDBYTES(x)						\
    do {							\
	int missing;						\
								\
	missing = (x) - bytes;					\
	if (missing > 0) {					\
	    if (header == head_buf) {				\
		if (missing <= end - buf) {			\
		    memcpy (header + bytes, buf, missing);	\
		    buf += missing;				\
		    bytes = (x);				\
		} else {					\
		    memcpy (header + bytes, buf, end - buf);	\
		    state_bytes = bytes + end - buf;		\
		    return 0;					\
		}						\
	    } else {						\
		memcpy (head_buf, header, bytes);		\
		state = DEMUX_HEADER;				\
		state_bytes = bytes;				\
		return 0;					\
	    }							\
	}							\
    } while (0)

#define DONEBYTES(x)		\
    do {			\
	if (header != head_buf)	\
	    buf = header + (x);	\
    } while (0)

    if (flags & DEMUX_PAYLOAD_START)
	goto payload_start;
    switch (state) {
    case DEMUX_HEADER:
	if (state_bytes > 0) {
	    header = head_buf;
	    bytes = state_bytes;
	    goto continue_header;
	}
	break;
    case DEMUX_DATA:
	if (demux_pid || (state_bytes > end - buf)) {
	    decode_mpeg2 (buf, end);
	    state_bytes -= end - buf;
	    return 0;
	}
	decode_mpeg2 (buf, buf + state_bytes);
	buf += state_bytes;
	break;
    case DEMUX_SKIP:
	if (demux_pid || (state_bytes > end - buf)) {
	    state_bytes -= end - buf;
	    return 0;
	}
	buf += state_bytes;
	break; 
    case DEMUX_MPA_DATA:
	if (demux_pid || (state_bytes > end - buf)) {
	    mpa_decode_data (buf, end, 0);
	    state_bytes -= end - buf;
	    return 0;
	}
	mpa_decode_data (buf, buf + state_bytes, 1);
	buf += state_bytes;
	break;
   }
   
    while (1) {
	if (demux_pid) {
	    state = DEMUX_SKIP;
	    return 0;
	}
    payload_start:
	header = buf;
	bytes = end - buf;
    continue_header:
	NEEDBYTES (4);
	if (header[0] || header[1] || (header[2] != 1)) {
	    if (demux_pid) {
		state = DEMUX_SKIP;
		return 0;
	    } else if (header != head_buf) {
		buf++;
		goto payload_start;
	    } else {
		header[0] = header[1];
		header[1] = header[2];
		header[2] = header[3];
		bytes = 3;
		goto continue_header;
	    }
	}  
	 if ((header[3] >=0xe0 && header[3]<= 0xef) && (demux_track == 0)){
	   demux_track = header[3];
	 }
	 switch (header[3]) {
	case 0xb9:	/* program end code */
	    /* DONEBYTES (4); */
	    /* break;         */
	    return 1;
	case 0xba:	/* pack header */
	    NEEDBYTES (5);
	    if ((header[4] & 0xc0) == 0x40) {	/* mpeg2 */
		NEEDBYTES (14);
		len = 14 + (header[13] & 7);
		NEEDBYTES (len);
		DONEBYTES (len);
		/* header points to the mpeg2 pack header */
	    } else if ((header[4] & 0xf0) == 0x20) {	/* mpeg1 */
		NEEDBYTES (12);
		DONEBYTES (12);
		/* header points to the mpeg1 pack header */
	    } else {
		fprintf (stderr, "weird pack header\n");
		DONEBYTES (5);
	    }
	    break;
	default:
	    if((header[3]>=0xc0) && (header[3]<=0xdf)){	// mpeg audio stream 
		NEEDBYTES (7);
		if ((header[6] & 0xc0) == 0x80) {	/* mpeg2 */
		    NEEDBYTES (9);
		    len = 9 + header[8];
		    NEEDBYTES (len);
		    /* header points to the mpeg2 pes header */
		    if (header[7] & 0x80) {
			uint32_t pts, dts;

			pts = (((header[9] >> 1) << 30) |
			       (header[10] << 22) | ((header[11] >> 1) << 15) |
			       (header[12] << 7) | (header[13] >> 1));
			dts = (!(header[7] & 0x40) ? pts :
			       (((header[14] >> 1) << 30) |
				(header[15] << 22) |
				((header[16] >> 1) << 15) |
				(header[17] << 7) | (header[18] >> 1)));
			apts = pts; adts=dts;
		    }
		} else {	/* mpeg1 */
		    int len_skip;
		    uint8_t * ptsbuf;

		    len = 7;
		    while (header[len - 1] == 0xff) {
			len++;
			NEEDBYTES (len);
			if (len >= 23) {
			    fprintf (stderr, "too much stuffing\n");
			    break;
			}	
		    }
		    if ((header[len - 1] & 0xc0) == 0x40) {
			len += 2;
			NEEDBYTES (len);
		    }
		    len_skip = len;
		    len += mpeg1_skip_table[header[len - 1] >> 4];
		    NEEDBYTES (len);
		    /* header points to the mpeg1 pes header */
		    ptsbuf = header + len_skip;
		    if ((ptsbuf[-1] & 0xe0) == 0x20) {
			uint32_t pts, dts;

			pts = (((ptsbuf[-1] >> 1) << 30) |
			       (ptsbuf[0] << 22) | ((ptsbuf[1] >> 1) << 15) |
			       (ptsbuf[2] << 7) | (ptsbuf[3] >> 1));
			dts = (((ptsbuf[-1] & 0xf0) != 0x30) ? pts :
			       (((ptsbuf[4] >> 1) << 30) |
				(ptsbuf[5] << 22) | ((ptsbuf[6] >> 1) << 15) |
				(ptsbuf[7] << 7) | (ptsbuf[18] >> 1)));

			apts = pts; adts=dts;
			
		    }
		}
		DONEBYTES (len);
		bytes = 6 + (header[4] << 8) + header[5] - len;
		if (demux_pid ||(bytes > end - buf)) {
		    mpa_decode_data (buf, end, 0);
		    state = DEMUX_MPA_DATA;
		    state_bytes = bytes - (end - buf);
		    return 0;
		} else if (bytes > 0) {
		    mpa_decode_data (buf, buf + bytes, 1);
		    buf += bytes;
		}
	    } //else if (header[3] >=0xe0 && header[3]<= 0xef){ /* video stream */
	    else if (header[3] == demux_track){ /* video stream */
//	    pes:
		NEEDBYTES (7);
		if ((header[6] & 0xc0) == 0x80) {	/* mpeg2 */
		    NEEDBYTES (9);
		    len = 9 + header[8];
		    NEEDBYTES (len);
		    /* header points to the mpeg2 pes header */
		    if (header[7] & 0x80) {
			uint32_t pts, dts;

			pts = (((header[9] >> 1) << 30) |
			       (header[10] << 22) | ((header[11] >> 1) << 15) |
			       (header[12] << 7) | (header[13] >> 1));
			dts = (!(header[7] & 0x40) ? pts :
			       (((header[14] >> 1) << 30) |
				(header[15] << 22) |
				((header[16] >> 1) << 15) |
				(header[17] << 7) | (header[18] >> 1)));
			mpeg2_tag_picture (mpeg2dec, pts, dts);

			vpts = pts; vdts=dts;
			
		    }
		} else {	/* mpeg1 */
		    int len_skip;
		    uint8_t * ptsbuf;

		    len = 7;
		    while (header[len - 1] == 0xff) {
			len++;
			NEEDBYTES (len);
			if (len > 23) {
			    fprintf (stderr, "too much stuffing\n");
			    break;
			}
		    }
		    if ((header[len - 1] & 0xc0) == 0x40) {
			len += 2;
			NEEDBYTES (len);
		    }
		    len_skip = len;
		    len += mpeg1_skip_table[header[len - 1] >> 4];
		    NEEDBYTES (len);
		    /* header points to the mpeg1 pes header */
		    ptsbuf = header + len_skip;
		    if ((ptsbuf[-1] & 0xe0) == 0x20) {
			uint32_t pts, dts;

			pts = (((ptsbuf[-1] >> 1) << 30) |
			       (ptsbuf[0] << 22) | ((ptsbuf[1] >> 1) << 15) |
			       (ptsbuf[2] << 7) | (ptsbuf[3] >> 1));
			dts = (((ptsbuf[-1] & 0xf0) != 0x30) ? pts :
			       (((ptsbuf[4] >> 1) << 30) |
				(ptsbuf[5] << 22) | ((ptsbuf[6] >> 1) << 15) |
				(ptsbuf[7] << 7) | (ptsbuf[18] >> 1)));
			mpeg2_tag_picture (mpeg2dec, pts, dts);

			vpts = pts; vdts=dts;
			
		    }
		}
		DONEBYTES (len);
		bytes = 6 + (header[4] << 8) + header[5] - len;
		if (demux_pid || (bytes > end - buf)) {
		    decode_mpeg2 (buf, end);
		    state = DEMUX_DATA;
		    state_bytes = bytes - (end - buf);
		    return 0;
		} else if (bytes > 0) {
		    decode_mpeg2 (buf, buf + bytes);
		    buf += bytes;
		}
	    } else if (header[3] < 0xb9) {
		fprintf (stderr,
			 "looks like a video stream, not system stream\n");
		DONEBYTES (4);
		return 2; /* Returning a positive value will allow us the chance
			to get other functions to quit trying to play an
			elementary (non-muxed) stream.  */
	    } else {
		NEEDBYTES (6);
		DONEBYTES (6);
		bytes = (header[4] << 8) + header[5];
//		skip:
		if (bytes > end - buf) {
		    state = DEMUX_SKIP;
		    state_bytes = bytes - (end - buf);
		    return 0;
		}
		buf += bytes;
	    }
	}
    }
}

/* Move all of this controller shit to another module.  Merge the control stuff for VCD
  and regular MPEG files into one generic interface. Should make code smaller and more
  easily readable. */
static int a_pressed;
static int b_pressed;
static int x_pressed;
static int y_pressed;
static int check_buttons()  /* Checks all of our buttons */
{
      MAPLE_FOREACH_BEGIN(MAPLE_FUNC_CONTROLLER, cont_state_t, st)
		if (!(st->buttons & CONT_A)) {
		    a_pressed = 0;
		    
		}
		if (!(st->buttons & CONT_B)) {
		    b_pressed = 0;
		    
		}
		if (!(st->buttons & CONT_X)) {
		    x_pressed = 0;
		    
		}
		if (!(st->buttons & CONT_Y)) {
		    y_pressed = 0;
		    
		}
		
		if (st->buttons & CONT_A) {
		    if (a_pressed==0)
		    {
		        a_pressed = 1;
			printf("Pressed A\n");
			return 1;
		    }
		} else if (st->buttons & CONT_Y) {
		    if (y_pressed==0)
		    {
		        y_pressed = 1;
			printf("Pressed Y\n");
			return 2;
		    }
		} else if (st->joyx > 90)/*(st->buttons & CONT_DPAD_RIGHT)*/ {
			printf("Pressed Right\n"); /* Start skipping frames like crazy. */
			seeking = 1;
			return 3;
		} else if (st->buttons & CONT_X) {
		    if (x_pressed==0)
		    {
		        x_pressed = 1;
			printf("Pressed x\n");
			return 4;
		    }
	      } else if (st->buttons & CONT_B) {
		    if (b_pressed==0)
		    {
		        b_pressed = 1;
			printf("Pressed B\n");
			return 5;
		    }
	      }
	MAPLE_FOREACH_END()
  return 0;
}

/*static*/ int pause_movie()
{
  int done;
  
  mpa_reset();
  vidclock_reset();
  done = 0;
  while (!(done==1))
  {
     done = check_buttons();
     if (done==3) done = 1;
  }

  return 1;
}

static void ps_loop (void)
{
    uint8_t * end;
    int stop_movie;
    int last_button;
    
    stop_movie = 0;

    do {
        last_button = check_buttons();
        switch(last_button)
	{
	  case 1:
	  if (seeking==1)
	  {  seeking = 0;  vidclock_reset();}
	  else
	  { pause_movie(); }
	   break;
	  case 2:
	   stop_movie = 1;
	   break;
	  case 3:
	   seeking = 1; mpa_reset(); vidclock_reset();
	   break;
	  default:
	   break;
	}
	end = vcd_buffer + fs_read (in_file, vcd_buffer, buffer_size);
	/*  define es_loop up above-- then, say, 'variable_name = (demux(buffer,end, 0);
	   Then, if we get a return greater than zero, still end, but then down below we can try
	   playing it as an elementary stream?  Or not... maybe give an option. */
	if (demux (vcd_buffer, end, 0))
	    break;	/* hit program_end_code */    
    } while ( (end == vcd_buffer + buffer_size && !sigint) && stop_movie == 0 );
    
}

static void es_loop (void)
{

    uint8_t * end;
    int stop_movie;
    int last_button;
    
    stop_movie = 0;

    do {
        last_button = check_buttons();
        switch(last_button)
	{
	  case 1:
	  if (seeking==1)
	  {  seeking = 0; vidclock_reset();}
	  else
	  { pause_movie(); }
	   break;
	  case 2:
	   stop_movie = 1;
	   break;
	  case 3:
	   seeking = 1;
	   break;
	  default:
	   break;
	}
	end = vcd_buffer + fs_read (in_file, vcd_buffer, buffer_size);
	decode_mpeg2 (vcd_buffer, end);
	if (stop_movie == 1)
	   break;
    } while (end == vcd_buffer + buffer_size && !sigint);

}

/* See if we can hook the EQ and stuff like that in here. */
static void mp3_loop (char * file_to_play)
{
    uint8_t * end;
    int ret, size;
    ID3 *info = NULL;
    struct mp3_audio_out *mp3_audio_out = &mp3_out;
    InitMP3(&mp); /* Open the MP3 decoder */
    int stop_movie = 0;
    int tag_at_start;
    
    /* This is probably not the best way to do this, but whatever. */
    in_file = fs_open (file_to_play, O_RDONLY);
    fs_read (in_file, vcd_buffer, 8192); 
    fs_close (in_file);
    
    if (vcd_buffer[0] == 'I' && vcd_buffer[1] == 'D' && vcd_buffer[2] == '3')
    { tag_at_start = 1; }else{ tag_at_start = 0;}
    /* Tags at the start of files can cause ear damage if we don't skip them. */
    
    info = create_ID3(info);
    
    parse_file_ID3(info, file_to_play);
    
    in_file = fs_open (file_to_play, O_RDONLY);
    if (tag_at_start==1)
    {
      if (info->tag_length) fs_seek(in_file, info->tag_length, SEEK_SET);
    }

    do {
    int last_button;
    last_button = check_buttons();
        switch(last_button)
	{
	  case 1:
	   pause_movie(); 
	   break;
	  case 2:
	   stop_movie = 1;
	   break;
	  default:
	   break;
	}
	
	draw_tag (info);
	
	end = vcd_buffer + fs_read (in_file, vcd_buffer, 8192);

	ret = decodeMP3_clipchoice(&mp,vcd_buffer,8192,mpa_out,&size,synth_1to1_mono, synth_1to1);
	if (mpeg_audio_activated==0)
	{
	/* Open an output driver */
   	 fo = mp3_audio_out->open(NULL, freqs[mp.fr.sampling_frequency], mp.fr.stereo, AFMT_S16_LE, 0, 8192);
	  mpeg_audio_activated = 1;
	}
	while(ret==MP3_OK)
    	{
          mp3_audio_out->write(fo,mpa_out,size);
	  ret = decodeMP3_clipchoice(&mp,NULL,0,mpa_out,&size,synth_1to1_mono, synth_1to1);	
    	}
	
	if (stop_movie) break;
    } while (end == vcd_buffer + 8192 && !sigint);
   
    mpeg_audio_activated = 0;
    mp3_audio_out->close(mp3_audio_out);
    ExitMP3(&mp); /* Close the MP3 decoder */
    fs_close (in_file);
    destroy_ID3(info);
}

int video_play(char * file_to_play) 
{
vo_driver_t const * drivers;

char *ext;

#ifdef HAVE_IO_H
    setmode (fileno (stdin), O_BINARY);
    setmode (fileno (stdout), O_BINARY);
#endif
   if (vcd_buffer==NULL)
   {
     vcd_buffer = (uint8_t *)malloc(2352*10);
   }
   ext = (char *)(strchr(file_to_play, '.')+1);
    
    demux_track = 0;//  0xe0;
    
    a_pressed = 1; /* ff_skip = 0; */
    
    adts = apts = vdts = vpts = 0;
    
    drivers = vo_drivers();
         
    output_open = drivers[0].open;
   
    output = output_open ();
    
    mpeg2dec = mpeg2_init ();
/* Looking at libmpeg2, I don't think we need to do this... */
  //  mpeg2_malloc_hooks (malloc_hook, NULL);

    mpeg_audio_activated = 0;
    if(!strcasecmp(ext,"mp1")||(!strcasecmp(ext,"mp2"))||(!strcasecmp(ext,"mp3")))
    {
        mp3_loop (file_to_play);
      
    }else{
    in_file = fs_open (file_to_play, O_RDONLY);
    bkg_clear(); /* The menu is using PVR memory that we need. */
    
    if(!strcasecmp(ext,"mpg")||(!strcasecmp(ext,"mpeg"))||(!strcasecmp(ext,"pss"))
    ||(!strcasecmp(ext,"sfd"))||(!strcasecmp(ext,"dat")))
	ps_loop ();
    if(!strcasecmp(ext,"m2v")||(!strcasecmp(ext,"m1v")))
        es_loop ();

	
    mpeg2_close (mpeg2dec);
    if (output->close)
	output->close (output);
	
    if (mpeg_audio_activated)
    {
      ExitMP3(&mp);
      audio_out->close(audio_out);
      mpeg_audio_activated = 0;
    }
	
    print_fps (1);
    bkg_reset(); /* Put our menu graphics back now that we don't need the space any longer. */
    fs_close (in_file);
    }
    return 0;
}

/* This does not handle pausing! It just deals with the
   fact that a pause was triggered, and shuts off audio
   and resets the fps counter. */
void mpegav_pause() 
{

    if (mpeg_audio_activated==1)
    {
      ExitMP3(&mp);
      audio_out->close(audio_out);
      mpeg_audio_activated = 0;
    }
    
    seeking = 0;
    vidclock_reset();   

}

/*
 * Below, we have our old 0.1/0.1a-ish VideoCD playback
 * routines (although I believe I changed them some just
 * before I rewrote all the VCD playback stuff in libvcd).
 * Bear in mind that this was done by a lot of trial and
 * error, and at that point in development was pretty
 * much a cheap hack until I got around to figuring out how
 * to parse the PSD file (eg, VCD/PSD.VCD or EXT/PSD_X.VCD).
 * All we do here is find the first sector of the second track
 * on the disc, the last sector on the disc, and play everything
 * in between.
 *
 * This was replaced by new code in our vcd library, which handles
 * playback control via Playlists and SelectionLists if present on the
 * disc, and has the ability to fake PBC on discs without it by using
 * the list of entry points we found upon opening the disc.
 *
 * There are a few functions below that are currently used by libvcd,
 * until I find a better way to allow it to interface with this module.
*/
CDROM_TOC *vcd_toc = NULL;

void vcd_reset() 
{
vo_driver_t const * drivers;
   
    demux_track = 0;// 0xe0;
    
    mpeg2_close (mpeg2dec);
    if (output->close)
	output->close (output);
	
    if (mpeg_audio_activated==1)
    {
      ExitMP3(&mp);
      audio_out->close(audio_out);
      mpeg_audio_activated = 0;
    }
    
    a_pressed = 1; b_pressed = 1; x_pressed = 1; y_pressed = 1;
    
    adts = apts = vdts = vpts = 0;

    seeking = 0;
    print_fps(1); vidclock_reset();
    
    drivers = vo_drivers();
         
    output_open = drivers[0].open;
   
    output = output_open ();
	
    
    mpeg2dec = mpeg2_init ();
    

}

#if 0
/* This is a friggin' mess!  This has been rewritten and the new implementation is in our libvcd. */
static int vcd_loop (_EntriesVcd_sim *entries_list, uint16 ch1)
{
    uint8_t * end;
    int stop_movie;
    int last_button;
    uint32 cur_sector;
    int next_entry, cur_entry, prev_entry;
    int i;
    int too_far;
    	
    stop_movie = 0;
    seeking = 0;
    last_button = 0;
    
    /* cur_sector = entries_list->ent_lba[0];
    next_entry = entries_list->ent_lba[1];
    cur_entry = 0;
    prev_entry = entries_list->ent_lba[0]; */
    
    cur_sector = entries_list->ent_lba[ch1];
    next_entry = entries_list->ent_lba[ch1+1];
    cur_entry = 0;
    prev_entry = (ch1 == 0) ? entries_list->ent_lba[0] : entries_list->ent_lba[ch1-1];
    
    /* I think we should actually subtract 150 from the LBA of the next track due to the pregap before each track,
      but there's no MPEG data in there so I don't think we're hurting anything here. */
    too_far = TOC_LBA(vcd_toc->entry[entries_list->track[entries_list->entry_count-1]]);
    
    if (too_far > TOC_LBA(vcd_toc->leadout_sector))
    {
      too_far = TOC_LBA(vcd_toc->leadout_sector);
    }
    
    do {

	last_button = check_buttons();
	if (last_button==1)  /* Pressed A (play/pause) */
	{
	  if (seeking==1) 
	  {  seeking = 0; vidclock_reset();}
	  else
	  { pause_movie(); }
	} else if (last_button==2){ /* Pressed Y (stop) */
	  stop_movie = 1;
	} else if (last_button==3){ /* Pressed right (ffwd) */
	  seeking = 1; mpa_reset(); vidclock_reset();
	} else if (last_button==4){ /* Pressed X (prev) */
	  cur_entry = (cur_entry > 0) ? cur_entry-1 : cur_entry;
	  cur_sector = prev_entry;
	  if (cur_entry)
	  {
	     prev_entry = entries_list->ent_lba[cur_entry-1];
	  }
	  else
	  {
	     prev_entry = prev_entry;//cur_sector;
	  }
	  next_entry = (cur_entry >= entries_list->entry_count) ? 
	  		0 : entries_list->ent_lba[cur_entry+1];
	  
	  adts = apts = vdts = vpts = 0;
	  mpa_reset();  vidclock_reset();
	  mpeg2_reset (mpeg2dec, 0);	 
				
	} else if (last_button==5){ /* Pressed B (next) */
	  if (cur_entry < entries_list->entry_count)
	   cur_entry++;
	   cur_sector = (next_entry==0) ? cur_sector : next_entry;
	   if(cur_entry)
	    {
	     prev_entry = entries_list->ent_lba[cur_entry-1];
	    }
	   else
	    {
	     prev_entry = prev_entry;
	    }
	  next_entry = (cur_entry >= entries_list->entry_count) ? 
	  		0 : entries_list->ent_lba[cur_entry + 1];
	    
	  adts = apts = vdts = vpts = 0;
	  mpa_reset(); vidclock_reset();
	  mpeg2_reset (mpeg2dec, 0);
	}
	
	cur_chapter = cur_entry;
	
	/* We have 24 bytes for sync and header/subheader
	  at the beginning of each sector, and 4 bytes
	  of extra data (error detection?) at the end.
	  I think.  We're supposed to have 2324 bytes
	  of usable data per sector, but we have 2352
	  per sector available.  This works. */
	
	end = vcd_buffer + 2348;
	 cdrom_read_sectors(vcd_buffer, cur_sector, 8); 
	cur_sector +=8; 
	
	if ( (next_entry > 0) && (cur_sector >= next_entry) )
	{
	  cur_entry++;
	  
	  prev_entry = entries_list->ent_lba[cur_entry-1];
	  next_entry = (cur_entry >= entries_list->entry_count) ? 
	  		0 : entries_list->ent_lba[cur_entry + 1];
	  if (cur_entry >= entries_list->entry_count) stop_movie = 1; 
	}
	
	for (i=0; i<8; i++)
	{
	   if (demux ( (vcd_buffer+(i*2352) )+24, (vcd_buffer+ (i*2352) )+2348, 0 ) )
	   { 
	    vcd_reset();
	    if (cur_entry == (entries_list->entry_count)-1) stop_movie = 1; 
	   }
	   if (stop_movie==1) break;
	}
	if (cur_sector >= too_far) stop_movie = 1;  
 
    } while ( (end == vcd_buffer + 2348 && !sigint) && stop_movie == 0 );
        
    return 0;
}
#endif

void vcd_setup_track() 
{
vo_driver_t const * drivers;
   
    demux_track = 0;//0xe0;
    
    a_pressed = 1; b_pressed = 1; x_pressed = 1; y_pressed = 1;
    
    adts = apts = vdts = vpts = 0;
    
    drivers = vo_drivers();
         
    output_open = drivers[0].open;
   
    output = output_open ();
    
    mpeg2dec = mpeg2_init ();
/* Looking at libmpeg2, I don't think we need to do this... */
   // mpeg2_malloc_hooks (malloc_hook, NULL); 

    mpeg_audio_activated = 0;
}

void vcd_close_track()
{
    mpeg2_close (mpeg2dec);
    if (output->close)
	output->close (output);
	
    if (mpeg_audio_activated == 1)
    {
      ExitMP3(&mp);
      audio_out->close(audio_out);
      mpeg_audio_activated = 0;
    }
	
    print_fps (1);

}

#if 0
int vcd_play(_EntriesVcd_sim *entries_list)
{
int track_2;
    if (vcd_toc==NULL)
    {
     vcd_toc = (CDROM_TOC *)malloc(sizeof(CDROM_TOC));
    }
  memset (vcd_toc, 0, sizeof(CDROM_TOC));
   cdrom_read_toc(vcd_toc, 0);
   
   bkg_clear(); /* The menu is using PVR memory that we need. */
    vcd_setup_track();
    vcd_loop(entries_list, 0);
    vcd_close_track();
   bkg_reset(); /* Put our menu graphics back now that we don't need the space any longer. */
  
  track_2 = TOC_LBA(vcd_toc->entry[entries_list->track[entries_list->entry_count-1]]);
  
// free (vcd_toc);
   return track_2;//1;
}
#endif
