/* GhettoPlay: an S3M browser and playback util
   (c)2000 Dan Potter
*/

#include <kos.h>
#include <string.h>
#include "gp.h"
#include "mpegav.h"
#include "vmu_controls.h"
#include "mp3junk.h"
#include "libvcd.h"
/* Takes care of the song menu */


/* Song menu choices */
typedef struct {
	char	fn[256];
	int	size;
} entry;
typedef struct {
	char	fn[256];
	int	size;
} lst_entry;

///////// Mutex protected /////////
static mutex_t * mut = NULL;

char curdir[1024] = "/cd";
char playdir[1024] = "/cd";
char loadme[256] = "";
char workstring[256] ="";

static entry entries[200];
/*static */int num_entries = 0;
//////// Mutex protected //////////
/*static */volatile int load_queued = 0;

/*static */int selected = 0, top = 0, iplaying = 0;

static int framecnt = 0;
static float throb = 0.2f, dthrob = 0.01f;

/* Here we hold the playlist */
/*static*/ int lst_size = 0;
static lst_entry lst_entries[200];
static int lst_playing = -1;


/* Code to draw a nice o-scope on the background during playback :) */
static mutex_t * hookmut = NULL;

static void load_song_list(/*void * p*/) {
	file_t d;

	d = fs_open(curdir, O_RDONLY | O_DIR);
	if (!d) {
		strcpy(curdir, "/");
		d = fs_open(curdir, O_RDONLY | O_DIR);
		if (!d) {
			mutex_lock(mut);
			num_entries = 1;
			strcpy(entries[0].fn,"Error!");
			entries[0].size = 0;
			mutex_unlock(mut);
			return;
		}
	}
	{
		dirent_t *de;
		num_entries = 0;
		if (strcmp(curdir, "/")) {
			mutex_lock(mut);
			strcpy(entries[0].fn, "<..>"); entries[0].size = -1;
			num_entries++;
			mutex_unlock(mut);
		}
		while ( (de = fs_readdir(d)) && num_entries < 200) {
		char *ext;
		ext = (char *)(strchr(de->name, '.')+1);
			printf("read entry '%s'\n", de->name);
			if (de->name[0] != '.') {
				mutex_lock(mut);
			/*if( (!stricmp(ext,"mpg"))||(!stricmp(ext,"mpeg"))||(!stricmp(ext,"pss"))
    			||(!stricmp(ext,"sfd"))||(!stricmp(ext,"m2v"))||(!stricmp(ext,"m1v"))
      			||(!stricmp(ext,"mp1"))||(!stricmp(ext,"mp2"))||(!stricmp(ext,"mp3")) )
       			{*/
				strcpy(entries[num_entries].fn, de->name);
				entries[num_entries].size = de->size;
				num_entries++;
			/*}*/
				mutex_unlock(mut);
			}
		}
	}
	fs_close(d);
	load_queued = 0;
}

/* Draws the song listing */
static void draw_listing() {
	float y = 110.0f;
	int i, esel;
	char fs_string[256];
	
//	printf("DEBUG: List size %d\r\n",lst_size);

	/* Draw all the song titles */	
//	for (i=0; i<14 && (top+i)<num_entries; i++) {
	for (i=0; i<10 && (top+i)<num_entries; i++) {
		int li;
		draw_poly_strf(60.0f, y, 100.0f, 1.0f, 1.0f, 1.0f, 1.0f,
			"%s", entries[top+i].fn);
		if (entries[top+i].size >= 0) {
			/* draw_poly_strf(32.0f+240.0f, y, 100.0f, 1.0f, 1.0f, 1.0f, 1.0f,
				"%d bytes", entries[top+i].size); */
		      if (entries[top+i].size >= 1024){				
			sprintf(fs_string, "%d KB", entries[top+i].size/1024);
		      }else{
		        sprintf(fs_string, "%d B", entries[top+i].size);
		      }
			draw_poly_strf( (544.0f - (((float)strlen(fs_string))*12.0f)), y, 100.0f, 1.0f, 1.0f, 1.0f, 1.0f,
				"%s", fs_string);
		} else {
			draw_poly_strf(400.0f, y, 100.0f, 1.0f, 1.0f, 1.0f, 1.0f, "<DIR>");
		}

		/* Check for playlist entries */
		for(li = 0; li<lst_size;li++)
		{
			strcat(workstring,curdir);
			strcat(workstring,"/");
			strcat(workstring,entries[top+i].fn);
	
//			printf("%s =? %s \r\n",workstring,lst_entries[li].fn);

			if(!(strcmp(workstring,lst_entries[li].fn)))
			{
				draw_poly_strf(568.0f, y, 100.0f, 1.0f, 1.0f, 1.0f, 1.0f, "%2d", li+1);
			}				
			if(!(strcmp(workstring,lst_entries[lst_playing].fn)))
			{
				draw_poly_box  (31.0f, y - 1.0f,
						609.0f, y + 25.0f, 90.0f,
						throb, 0.0f, 0.2f, 0.2f, 
						throb, 0.0f, throb, 0.2f);

			}
			
			workstring[0]=0;
		}		
		
		
		y += 24.0f;
	}
	
	/* Put a highlight bar under one of them */
	esel = (selected - top);
	/*draw_poly_box(31.0f, 92.0f+esel*24.0f - 1.0f,
		609.0f, 92.0f+esel*24.0f + 25.0f, 95.0f,
		throb, throb, 0.2f, 0.2f, throb, throb, 0.2f, 0.2f); */
	draw_poly_box(60.0f, 110.0f+esel*24.0f - 1.0f,
		580.0f, 110.0f+esel*24.0f + 25.0f, 95.0f,
		throb, 0.2f, 0.2f, throb, throb, 0.2f, 0.2f, throb);

}

static void start(char *fn) {
	  vmu_set_icon(controls_xpm);
	  video_play(fn);
}

static void stop() {

}

/* Handle controller input */
/* static uint8 mcont = 0; */

static void set_eq_menu() /* Set our 32-band graphic equalizer */
{
  int i, in_menu = 1, band = 0, up_moved = 0, down_moved = 0, left_moved = 0, right_moved = 0,
		  a_pressed = 0;
  static int band_value[32] = {
  100, 100, 100, 100, 100, 100, 100, 100,
  100, 100, 100, 100, 100, 100, 100, 100,
  100, 100, 100, 100, 100, 100, 100, 100,
  100, 100, 100, 100, 100, 100, 100, 100
  };
  static maple_device_t * mcont = NULL;
  static cont_state_t * state = NULL;

  if (!mcont)   mcont = maple_enum_type(0, MAPLE_FUNC_CONTROLLER);



   for (i=0; i< 32; i++)    band_value[i] = get_eq_band(i);

   a_pressed = framecnt;
  while (in_menu==1)
  {
    state = (cont_state_t *)maple_dev_status(mcont);
    if (!state) return;

    if ((state->buttons & CONT_DPAD_UP))
    {
     if ((framecnt - up_moved) > 3)
     {
      band_value[band]++;
      if (band_value[band] > 200) band_value[band] = 200;
      set_eq_band(band, band_value[band]);
      up_moved = framecnt;
     }
    }
    if ((state->buttons & CONT_DPAD_DOWN))
    {
     if ((framecnt - down_moved) > 3)
     {
      band_value[band]--;
      if (band_value[band] < 0 ) band_value[band] = 0 ;
      set_eq_band(band, band_value[band]);
      down_moved = framecnt;
     }
    }
    if ((state->buttons & CONT_DPAD_LEFT))
    {
      if ((framecnt - left_moved) > 10)
      {
        band--; if (band <0) band = 0;
	left_moved = framecnt;
      }
    }
    if ((state->buttons & CONT_DPAD_RIGHT))
    {
      if ((framecnt - right_moved) > 10)
      {
        band++; if (band > 31) band = 31;
	right_moved = framecnt;
      }
    }
    if ((state->buttons & CONT_A))
    {
      if((framecnt - a_pressed) > 10)
      {
	in_menu = 0;
      }
    }
    if ((state->buttons & CONT_X))
    {
      for (i=0; i<32; i++)
      {
       band_value[i]= 100;
       set_eq_band(i, 100);
      }
    }
	     
	pvr_wait_ready();
	pvr_scene_begin();
	pvr_list_begin(PVR_LIST_OP_POLY);

	/* Opaque list *************************************/
	bkg_render();

	/* End of opaque list */
	pvr_list_finish();
	pvr_list_begin(PVR_LIST_TR_POLY);

	/* Translucent list ********************************/
	draw_poly_strf(5.0f, 60.0f, 100.0f, 1.0f, 1.0f, 1.0f, 1.0f,
			//"         VC/DC  Pre-0.2 Alpha (C)2006-2008         ");
			  "     Press X to reset EQ or A to save and exit.    ");
	
	draw_poly_box(60.0f, 240.0f, 580.0f, 241.0f, 75.0f,
		1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f);	
	for (i=0; i< 32; i++)
	{
	  draw_poly_box(160.0f + i*10.0f, 140.0f,
		161.0f + i*10.0f, 340.0f, 75.0f,
		1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f);
	 /* Draw the fader control */
	  draw_poly_box(156.0f + i*10.0f, 340.0f - (float)(band_value[i]) - 2.0f,
		164.0f + i*10.0f + 4.0f, 340.0f - (float)(band_value[i]) + 2.0f, 75.0f,
		1.0f, 1.0f, 0.0f, 0.0f, 1.0f, 1.0f, 0.0f, 0.0f);
	}			
        draw_poly_box(155.0f + band*10.0f, 140.0f,
		165.0f + band*10.0f, 340.0f, 95.0f,
		throb, 0.2f, 0.2f, throb, throb, 0.2f, 0.2f, throb);						
	pvr_list_finish();
	pvr_scene_finish();
	
        
	framecnt++;
	/* Adjust the throbber */
	throb += dthrob;
	if (throb < 0.2f || throb > 0.8f) {
		dthrob = -dthrob;
		throb += dthrob;
	}
  }
}

#define TOTAL_MENU_ITEMS 4
extern int pbc_switch;
/* This would probably make more sense with some sort of a struct for the menu items, but whatever. */
static void settings_menu()
{
	static int up_moved = 0, down_moved = 0, left_moved = 0, right_moved = 0,
		  a_pressed = 0, in_menu = 1, selection = 0, menu_item = 0;
	static int last_selection[TOTAL_MENU_ITEMS];
	char *optionlist[TOTAL_MENU_ITEMS][3]=
	{
	  {"Best audio (default)", "Okay audio", "Worst audio"}, /* Audio Quality */
	  {"Scoop (default)"," Bass Boost",  "Treble Reduce"},  /* EQ Preset*/
	  {"Off", "On","Set"},					/* EQ on/off */
	  {"On(Default)", "Off"," "}				/* PBC on/off */
	};

  static maple_device_t * mcont = NULL;
  static cont_state_t * state = NULL;

  if (!mcont)   mcont = maple_enum_type(0, MAPLE_FUNC_CONTROLLER);

     in_menu = 1;
     
     a_pressed = framecnt;
     
     while (in_menu==1)
     {	
    state = (cont_state_t *)maple_dev_status(mcont);
    if (!state) return;
	if ((state->buttons & CONT_DPAD_UP)) {
		if ((framecnt - up_moved) > 10) {
			if (menu_item > 0) {
				last_selection[menu_item] = selection;
				menu_item--;
				selection = last_selection[menu_item];
			}
			up_moved = framecnt;
		}
	}
	if ((state->buttons & CONT_DPAD_DOWN)) {
		if ((framecnt - down_moved) > 10) {
			if (menu_item < TOTAL_MENU_ITEMS-1) {
				last_selection[menu_item] = selection;
				menu_item++;
				selection = last_selection[menu_item];
			}
			down_moved = framecnt;
		}
	}
	
	if ((state->buttons & CONT_DPAD_LEFT)) {
		if ((framecnt - left_moved) > 10) {
			if (selection >= 0) {
				selection--;
				if (selection < 0) {
					selection = 2;
				}
			}
			left_moved = framecnt;
			if (menu_item==0) set_audio_quality(selection+1);
			if (menu_item==1) set_eq(selection);
			if ((menu_item==2) && (selection < 2) )enable_eq(selection);
			if ((menu_item==3) && (selection < 2) )pbc_switch = 1-selection;
		}
	}
	if ((state->buttons & CONT_DPAD_RIGHT)) {
		if ((framecnt - right_moved) > 10) {
			if (selection <= 2) {
				selection++;
				if (selection > 2) {
					selection = 0;
				}
			}
			right_moved = framecnt;
			if (menu_item==0) set_audio_quality(selection+1);
			if (menu_item==1) set_eq(selection);
			if ((menu_item==2) && (selection < 2) )enable_eq(selection);
			if ((menu_item==3) && (selection < 2) )pbc_switch = 1-selection;
		}
	}
	if ((state->buttons & CONT_A)){
               if((framecnt - a_pressed) > 10) {
	       in_menu = 0;
	       
	       }
	}
	     
	pvr_wait_ready();
	pvr_scene_begin();
	pvr_list_begin(PVR_LIST_OP_POLY);

	/* Opaque list *************************************/
	bkg_render();

	/* End of opaque list */
	pvr_list_finish();
	pvr_list_begin(PVR_LIST_TR_POLY);

	/* Translucent list ********************************/
	draw_poly_strf(60.0f, 110.0f, 100.0f, 1.0f, 1.0f, 1.0f, 1.0f,
			"Audio Setting:");
	draw_poly_strf(60.0f, 134.0f, 100.0f, 1.0f, 1.0f, 1.0f, 1.0f,
			"MPEG EQ Setting:");
	draw_poly_strf(60.0f, 158.0f, 100.0f, 1.0f, 1.0f, 1.0f, 1.0f,
			"MPEG EQ:");
	draw_poly_strf(60.0f, 182.0f, 100.0f, 1.0f, 1.0f, 1.0f, 1.0f,
			"PBC:");						
	draw_poly_strf( (568.0f - (((float)strlen(optionlist[menu_item][selection]))*12.0f)), 110.0f+menu_item*24.0f, 100.0f, 1.0f, 1.0f, 1.0f, 1.0f,
				"%s",optionlist[menu_item][selection]);	
				
        draw_poly_box(60.0f, 110.0f+menu_item*24.0f - 1.0f,
		580.0f, 110.0f+menu_item*24.0f + 25.0f, 95.0f,
		throb, 0.2f, 0.2f, throb, throb, 0.2f, 0.2f, throb);						
	pvr_list_finish();
	pvr_scene_finish();
	
        
	framecnt++;
	/* Adjust the throbber */
	throb += dthrob;
	if (throb < 0.2f || throb > 0.8f) {
		dthrob = -dthrob;
		throb += dthrob;
	}	
     
     }
    if(menu_item==2 && selection==2)set_eq_menu();
}
void check_controller() {
	static int up_moved = 0, down_moved = 0, a_pressed = 0, y_pressed = 0, x_pressed = 0;
  static maple_device_t * mcont = NULL;
  static cont_state_t * state = NULL;

  if (!mcont)   mcont = maple_enum_type(0, MAPLE_FUNC_CONTROLLER);

    state = (cont_state_t *)maple_dev_status(mcont);
    if (!state) return;

	if ((state->buttons & CONT_DPAD_UP)) {
		if ((framecnt - up_moved) > 10) {
			if (selected > 0) {
				selected--;
				if (selected < top) {
					top = selected;
				}
			}
			up_moved = framecnt;
		}
	}
	if ((state->buttons & CONT_DPAD_DOWN)) {
		if ((framecnt - down_moved) > 10) {
			if (selected < (num_entries - 1)) {
				selected++;
//				if (selected >= (top+14)) {
				if (selected >= (top+10)) {
					top++;
				}
			}
			down_moved = framecnt;
		}
	}
	if (state->ltrig > 0) {
		if ((framecnt - up_moved) > 10) {
//			selected -= 14;
			selected -= 10;

			if (selected < 0) selected = 0;
			if (selected < top) top = selected;
			up_moved = framecnt;
		}
	}
	if (state->rtrig > 0) {
		if ((framecnt - down_moved) > 10) {
//			selected += 14;
			selected += 10;
			if (selected > (num_entries - 1))
				selected = num_entries - 1;
//			if (selected >= (top+14))
			if (selected >= (top+10))
				top = selected;
			down_moved = framecnt;
		}
	}

	if ((state->buttons & CONT_Y)) {
		if ((framecnt - y_pressed) > 10)
		{
		strcat(workstring,curdir);
		strcat(workstring,"/");
		strcat(workstring,entries[selected].fn);
		strcpy(lst_entries[lst_size].fn,workstring);
		
		lst_entries[lst_size].size = entries[selected].size;
		printf("DEBUG: Entry %d : %s\r\n",lst_size,lst_entries[lst_size].fn);

		workstring[0]=0;
		
		lst_size++;

		}
		y_pressed=framecnt;
	}
	
	if ((state->buttons & CONT_X) ) {
	     if ((framecnt - x_pressed) > 10){
		/* End of translucent list */
		pvr_list_finish();

		/* Finish the frame *******************************/
		pvr_scene_finish();
		settings_menu();
		a_pressed = framecnt;
		x_pressed = framecnt;
		
		/* Restart a scene--we're in the middle of one before here, so we should set one back up. */
		pvr_wait_ready();
		pvr_scene_begin();
		pvr_list_begin(PVR_LIST_OP_POLY);

		/* Opaque list *************************************/
		bkg_render();

		/* End of opaque list */
		pvr_list_finish();
		pvr_list_begin(PVR_LIST_TR_POLY);
		
	    }
	}
	if ((state->buttons & CONT_A) && !load_queued) {
		if ((framecnt - a_pressed) > 10)
		{
		  /* End of translucent list */
		  pvr_list_finish();

		  /* Finish the frame *******************************/
		  pvr_scene_finish();
		
			if (!strcmp(entries[selected].fn, "Error!"))
			{
				num_entries = 0;
			}
			else if (lst_size > 0)
			{
				printf("DEBUG: Playing playlist...\r\n");
				lst_playing=0;
				stop();
				for (lst_playing=0;lst_playing < lst_size; lst_playing++)
				{
				  if (lst_entries[lst_playing].size > 0)
				  start(lst_entries[lst_playing].fn);
				  thd_sleep(100);
				}
				lst_size = lst_playing = 0;
				y_pressed=framecnt; /* We do this because in-movie, 'Y' is stop. */
				x_pressed=framecnt;
			}
			else if (entries[selected].size >= 0)
			{
			  char *ext;
				stop();
				strcpy(playdir,curdir);
				strcat(loadme,curdir);
				strcat(loadme,"/");
				strcat(loadme,entries[selected].fn);
				
				ext = (char *)(strchr(entries[selected].fn, '.')+1);
				if(!strcasecmp(ext,"bin"))
				{
				  vmu_set_icon(vcdcontrols_xpm);
				  play_vcd_image(loadme);
				}else{
				 start(loadme);
				}
				loadme[0]=0;
				iplaying=selected;
				y_pressed=framecnt; /* We do this because in-movie, 'Y' is stop. */
				x_pressed=framecnt;
			}
			else
			{
				if (!strcmp(entries[selected].fn, "<..>"))
				{
					int i;
					for (i=strlen(curdir); i>0; i--)
					{
						if (curdir[i] == '/')
						{
							curdir[i] = 0;
							break;
						}
						else
						{
							curdir[i] = 0;
						}
					}
				} 
				else
				{
					if (strcmp(curdir, "/"))
					strcat(curdir, "/");
					strcat(curdir, entries[selected].fn);
				}
				mutex_lock(mut);
				selected = top = num_entries = 0;
				mutex_unlock(mut);
				printf("current directory is now '%s'\n", curdir);
			}
		pvr_wait_ready();
		pvr_scene_begin();
		pvr_list_begin(PVR_LIST_OP_POLY);

		/* Opaque list *************************************/
		bkg_render();

		/* End of opaque list */
		pvr_list_finish();
		pvr_list_begin(PVR_LIST_TR_POLY);

		/* Translucent list ********************************/
		draw_listing();
		pvr_list_finish();
		pvr_scene_finish();
			
		}
		a_pressed = framecnt;
	}
}

/* Check maple bus inputs */
void check_inputs() {
	check_controller();
}

/* Main rendering of the song menu */
void song_menu_render() {
	if (mut == NULL) {
		mut = mutex_create();
		hookmut = mutex_create();
	}

	/* Draw a background box */
	/*draw_poly_box(30.0f, 80.0f, 610.0f, 440.0f-96.0f, 90.0f, 
		0.2f, 0.8f, 0.5f, 0.0f, 0.2f, 0.8f, 0.8f, 0.2f); */
	/* draw_poly_box(53.0f, 110.0f, 582.0f, 350.0f, 90.0f, 
		0.75f, 0.2f, 0.2f, 0.2f, 0.75f, 0.2f, 0.2f, 0.2f); */
		
	/* If we don't have a file listing, get it now */
	if (num_entries == 0 && !load_queued) {
		load_queued = 1;
		//thd_create(0, load_song_list, NULL);
		load_song_list();
	}

	/* if (load_queued)
		draw_poly_strf(32.0f, 82.0f, 100.0f, 1.0f, 1.0f, 1.0f, 1.0f,
			"Scanning Directory..."); */
	
	/* Draw the song listing */
	mutex_lock(mut);
	draw_listing();
	mutex_unlock(mut);
	
	/* Adjust the throbber */
	throb += dthrob;
	if (throb < 0.2f || throb > 0.8f) {
		dthrob = -dthrob;
		throb += dthrob;
	}
	
	/* Check maple inputs */
	check_inputs();
	
	framecnt++;
}


