/*  libvcd.c Copyright (c) Josh Sutherland.
 * Copyright (C) 2008 Josh Sutherland <OneThirty8@aol.com>
 * This file is part of VC/DC, a free MPEG-2 video player.
 *
 * VC/DC is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * VC/DC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/* This code is based on vcdimager-0.7.23 by Herbert Valerio Riedel, 
 * Jens B. Jorgensen, and Rocky Bernstein (http://www.vcdimager.org)
 * as well as the xine-vcdnav plugin by Rocky Bernstein
 * (http://sf.net/projects/xine -- xine-vcdnav was the current CVS version
 * on or about 24 June, 2006).  libcdio-0.77 was used for several things
 * also (http://www.gnu.org/software/libcdio), although obviously not for
 * the actual reading of Compact Discs on the Dreamcast.
 *
 */
#include <kos.h>
#include <sys/time.h>
#include <libvcd.h>
#include <mpegav.h> 
#include "gettimeofday.h"
#include "gp.h"

#include "vcdbuttons.h"
#include "inttypes.h"
#include "vcd_iso.h"

int (* read_sectors)(void *buffer, int sector, int cnt);
int (* read_toc)(CDROM_TOC *toc_buffer, int session);

/* Stuff I should have in a header file somewhere... */
extern int vcd_play_chapter(_EntriesVcd_sim *entries_list, uint16 chapter);
extern int vcd_play_still(int chapter_lba);
extern int vcd_play_track(int track_id);
extern int vcd_dummy_poll();
extern void vcd_setup_track();
extern void vcd_close_track();

extern void redraw_frame();

extern int pbc_get_button();
extern int pbc_handle_button();
extern uint32 cur_sector;

extern uint8_t *vcd_buffer;
extern CDROM_TOC *vcd_toc;

extern void bkg_clear();
extern void bkg_reset();

int pbc_switch = 1; /* So the user can turn PBC on and off. */
/*
  Private prototypes and whatnot. 
*/

/* This is probably not named very properly--this will simply contain enough information for us to determine
   if a remote control button entered by the user via the Dreamcast controller/onscreen numerical selection
   is, in fact, a valid button entry for the given SelectionList/PlayList, and give us information about the
   contents of a SelectionList to alter behavior (such as enabling hot-spots and whatnot).*/

RemoteControl_t * remotecontrol = NULL;

/* A pointless container for our filenames/filepaths. For SVCD/HQ-VCD, we're only going to read one
   of SCANDATA.DAT or SEARCH.DAT, since all I care about are the I-frame addresses. */
static struct {
/* VCD/SVCD folder */
 char entries[22]; /* /vcd/svcd/entries.svd + NULL byte = 22 bytes. */
 char psd[18];
 char lot[18]; /* /vcd/svcd/lot.svd + NULL byte = 18 bytes. */
/* char tracks[21]; */ /* Probably don't need this... */
 char search[21]; /* This is almost identical to the VCD 2.0 SCANDATA.DAT... */
/* EXT Folder */
 char psd_x[19]; /* Omit these?  We could just point at these if we have Extended PBC available. */
 char lot_x[19];  /* /vcd/ext/lot_x.vcd + NULL byte = bytes. */
 char scandata[22];  /* /vcd/ext/scandata.dat + NULL byte = 22 bytes. */  
} VCD_names;

/* These are raw copies of the INFO.VCD, ENTRIES.VCD, LOT.VCD, and PSD.VCD files. All but one will
   let us read one or more cdrom sectors into it and magically fill a struct with useful information.*/
   
static _InfoVcd_t *infovcd = NULL;
static _EntriesVcd_t *entriesvcd = NULL;
/*static*/ _EntriesVcd_sim *simple_entries = NULL;
static _LotVcd_t *lotvcd = NULL;  /* I don't even know if we need to mess with the List Offset Table, since the info we need
		    should be present in PSD.VCD (it contains the offsets as well) but this is supposed to allow
		    us to access the disc from offsets other than the default (which I assume is offset 0) so
		    we'll see if I feel like making that feature available and leave this here for now. */
		    
static char *psdvcd = NULL;  /* Unlike most of the other files in the first track, PSD.VCD doesn't have a simple
		format, but we'll be able to make sense of this raw data when it's time to use it. We'll get an
		offset into this buffer, so the byte psdvcd[offset] will tell us what type of struct to fill in.
		We will get an endlist descriptor, a playlist descriptor or a selection list descriptor that contains
		information on what to play (if anything) and where to go if user presses 'next' or 'previous' or '1' or whatever.
		There's also a PSD Command List, but xine doesn't seem to use it... although if you want to talk
		about digging through code, look at either of the VCD plugins in the xine CVS on sourceforge.
		I had to dig through the plugin, all of vcdimager, and all of libcdio and I'm sure I missed some things.
		We're lucky that the Dreamcast doesn't have zillions of variables to consider like a modern PC does.
		*/

static ScandataDat_v2_t *scandata = NULL; /* /EXT/SCANDATA.DAT is an optional file in VCD 2.0 and SVCD that contains the addresses of
					the I-frames, which is very useful for seeking, and I'm going to try to use it to allow
					us to play backwards.  Fingers crossed... */
static SearchDat_t *search_dat = NULL; /* Almost the same as the VCD 2.0 version of Scandata.dat. */

static int first_seg_addr = 0;  /* This is stored in INFO.VCD as an msf struct--we want the LBA. */

/* So we don't have to type such long names.  We'll be typecasting a lot. */
typedef _PsdPlayListDescriptor_t _PlayList;
typedef _PsdSelectionListDescriptor_t _SelectionList;
typedef PsdSelectionListDescriptorExtended_t _ExtList;

static _PlayList *PlayList;
static _SelectionList *SelectionList;
static _ExtList *ExtList;
/*
  VCD reading functions
*/
static int is_vcd();  /* So we can prevent trying to play a disc that isn't there anymore. The fuction is down below somewhere...*/

/*static int Pbc_PlaySpi(uint16 play_item) :
  We should add a length thing in here to handle continuations.
  Basically, I think that if we start with the desired play item and loop through the rest,
  stopping when we find one that is not a continuation, we can get the number of 'segment chunks'
  the desired play item actually contains.  Multiply that by 150 if we want the number of sectors.
*/
static int Pbc_PlaySpi(uint16 play_item)
{
   int i;
   play_item = play_item - 1000;

   vcd_setup_track();
   
   do{
     i = vcd_play_still(first_seg_addr + (150 * play_item));
     play_item++;
   } while (infovcd->spi_contents[play_item].item_cont == true);
   
   vcd_close_track();
   return i;
}


/* static int Pbc_PlayTrack(uint16 play_item) :
   We should actually be able to just use the track number here instead of filling in a whole list of entrypoints.
*/
static int Pbc_PlayTrack(uint16 play_item)
{
  int i;
  
  i = vcd_play_track(play_item);
 
 return i;
}

/* Pbc_PlayEntry(uint16 play_item) :
   This is fairly simple.  Play from entry[play_item-100] to the end of that track.
*/
static int Pbc_PlayEntry(uint16 play_item)
{
  int i;
  
  i = vcd_play_chapter(simple_entries, play_item-100 );
   
   return i;  
}
/*  This is already defined in libvcd.h.  I just copied it here to make writing the Pbc_SelectionList function easier.

typedef struct _PsdSelectionListDescriptor_tag {
  uint8 type;
  PsdSelectionListFlags_t flags;
  uint8 nos;
  uint8 bsn;
  uint16 lid;
  uint16 prev_ofs;
  uint16 next_ofs;
  uint16 return_ofs;
  uint16 default_ofs;
  uint16 timeout_ofs;
  uint8 totime;
  uint8 loop;
  uint16 itemid;
  uint16 ofs[EMPTY_ARRAY_SIZE]; 
}
typedef struct
  {
    uint8 x1; < upper left 
    uint8 y1; < upper left 
    uint8 x2; < lower right 
    uint8 y2; < lower right 
} GNUC_PACKED psd_area_t;

typedef struct {
  psd_area_t prev_area;
  psd_area_t next_area;
  psd_area_t return_area;
  psd_area_t default_area;
  psd_area_t area[EMPTY_ARRAY_SIZE];  variable length 
} GNUC_PACKED PsdSelectionListDescriptorExtended_t;
*/
static uint16 Pbc_SelectionList(char *selection_list)
{
int i; /* A multi-purpose integer. */
 uint16 *offsets; /* Where to go if we get a numeric, next, return, etc button. */
 uint16 prev_ofs, next_ofs, return_ofs, default_ofs, timeout_ofs;
 uint16 itemid;
 uint8 loop;
 uint16 our_offset;
 uint64 then = 0, now = 0;
 int control_value;
 
   /* We can't convert the original copy to LE because we might be handed that
      PSD offset again, so we make a temporary copy of the 16-bit integers here and do our work based on those. */
   SelectionList = (_SelectionList *)selection_list;
 
   offsets = (uint16 *)malloc(2*SelectionList->nos);
   
   memcpy (offsets, selection_list + 20, 2*SelectionList->nos);
   
   /* If we hand this stuff to the remote control, we can get the area stuff handled instead of the buttons. 
      But maybe we should really check the flags for this... (In vcdimager, it appears that the flag is only
      set on SVCD, but still... check for it. */
   if ( (SelectionList->type == PSD_TYPE_EXT_SELECTION_LIST) || (SelectionList->flags.SelectionAreaFlag == true) )
   {
     ExtList = (_ExtList *)(selection_list + 20 + (2*SelectionList->nos));
     remotecontrol->ExtList = ExtList;
   }else{
     remotecontrol->ExtList = NULL;
   }
   
   prev_ofs = SelectionList->prev_ofs;
   next_ofs = SelectionList->next_ofs;
   return_ofs = SelectionList->return_ofs;
   default_ofs = SelectionList->default_ofs;
   timeout_ofs = SelectionList->timeout_ofs;
   itemid = SelectionList->itemid;
 
   /* Byteswap our copy. */
   prev_ofs = UINT16_SWAP_LE_BE_C(prev_ofs);
   next_ofs = UINT16_SWAP_LE_BE_C(next_ofs);
   return_ofs = UINT16_SWAP_LE_BE_C(return_ofs);
   default_ofs = UINT16_SWAP_LE_BE_C(default_ofs);
   timeout_ofs = UINT16_SWAP_LE_BE_C(timeout_ofs);
   itemid = UINT16_SWAP_LE_BE_C(itemid);
 
   for (i=0; i < SelectionList->nos; i++)
   {
     offsets[i] = UINT16_SWAP_LE_BE_C(offsets[i]);
   }
 
   our_offset = 0xffff;
   loop = SelectionList->loop;
   if (!loop) loop = 1; /* Wrong!  if (loop == 0) we loop indefinitely! We will handle this below. */
   
   /* Fill in our RemoteControl_t struct with the values for this list. */
   remotecontrol->bsn = SelectionList->bsn;
   remotecontrol->nos = (remotecontrol->default_ofs == PSD_OFS_MULTI_DEF_NO_NUM) ? 0 : SelectionList->nos;
   remotecontrol->prev_ofs = prev_ofs;
   remotecontrol->next_ofs = next_ofs;
   remotecontrol->return_ofs = return_ofs;
   remotecontrol->default_ofs = default_ofs; 
   remotecontrol->ptime = 0x0000; /* Play full item--'ptime' is only specified in PlayList, not SelectionList. */
   remotecontrol->atime = 0; /* Autopause time not specified in SelectionList either. */ 
   remotecontrol->jump_delay = (SelectionList->loop & JUMP_DELAYED) ? 1 : 0; /* This must be handled in pbc.c */
   /* then = now = timer_ms_gettime64(); */ /* We need to figure a way to handle this below--see note about 'timeout time.'*/
   
   while (our_offset == 0xffff)  /* We have to allow for the user interaction buttons to give us the next offset. */
   {
     if (loop)
    { 
     if ( (itemid >1)     && (itemid < 100)  )
     { 
         control_value = Pbc_PlayTrack(itemid);
     }else if ( (itemid >=100)  && (itemid < 600)  )
     {
     	 control_value = Pbc_PlayEntry(itemid);
     }else if ( (itemid >=1000)  && (itemid < 2980)  )
     {
     	 control_value = Pbc_PlaySpi(itemid);
     } /* else, we're kinda screwed because we've got an invalid itemid. */
     else
     {
         control_value = vcd_dummy_poll();
     }
     then = now = timer_ms_gettime64(); /* Start our timer _AFTER_ the play-item is done, which should be correct. */
    } else { 
     control_value = vcd_dummy_poll();
     redraw_frame();
    }
    
    switch(control_value)
    {
     case BUTTON_PREV:
         our_offset = prev_ofs;
         break;
     case BUTTON_NEXT:
         our_offset = next_ofs;
         break;
     case BUTTON_RET:
         our_offset = return_ofs;
         break;
     case BUTTON_DEF:
         our_offset = default_ofs;
	 break;
     case BUTTON_NUMERIC:  /* Handle bsn in the button validation thing in pbc.c. */
         i = pbc_handle_button(); /*pbc_get_button();*/
	 if ( (i <= SelectionList->nos) && (i > 0) )
         {
	   our_offset = offsets[i - 1]; /* But our array is zero-based, so subtract 1. */
	 }
	 break;
     case BUTTON_STOP:
	our_offset = 0xffff;
	goto selection_end;
	break;
     default:
         break;
    }
    
     if (loop) loop--;
     
     if (SelectionList->loop == 0) loop = 1;
     
     /* Is this wrong? 'totime' = 'timeout time,' not 'total time,' so we want to start counting after the play-item plays, right? */
     if ((loop==0) && ( ((now-then)/1000) >= SelectionList->totime) && our_offset == 0xffff)
     {
       our_offset = timeout_ofs;
       if ( (our_offset == 0xffff) && (SelectionList->nos > 0) )
       {
         struct timeval tv;
	 
	 gettimeofday(&tv, NULL);
	 
	 srand(tv.tv_sec);
	 
	 if (SelectionList->nos > 0)
	 {
	   i = (1+rand())%SelectionList->nos;
	   our_offset = offsets[i-1];
	 }
       } /*end of if ( (our_offset == 0xffff) && (SelectionList->nos > 0) ) */
       
     }
     now = timer_ms_gettime64();
   }

selection_end:
/* Here we need to handle multi-default offsets, even if we do it incorrectly (I'm really just guessing
   how this feature is supposed to work) just so we don't go nowhere! */
  if ((our_offset == PSD_OFS_MULTI_DEF_NO_NUM) || (our_offset == PSD_OFS_MULTI_DEF))
  {
    int track;
    int entry_start_offset; /* How many entries to skip over. */
    int entry_max; /* The last entry point in the play item. */
    
    if ((itemid < 2) || (itemid > 600))
    {
      our_offset = PSD_OFS_DISABLED; /* Only works if itemid points to an MPEG sequence/entrypoint. */
      goto selection_end;
    }
    
    track =  (itemid >= 100) ? simple_entries->track[itemid-100] : itemid;
    for (i = 0; i < simple_entries->entry_count; i++)
    {
      if (simple_entries->track[i] == track)
      break;
    }
    
    entry_start_offset = i;
    for (i = simple_entries->entry_count - 1; i >=0; i--)
    {
      if (simple_entries->track[i] == track)
      break;
    }
    entry_max = i;
    
    do {
    entry_max--;
    }  while (simple_entries->ent_lba[entry_max] > cur_sector);
    
    entry_max -= entry_start_offset;
    our_offset = offsets[entry_max];
  } 
  free (offsets);
  return our_offset;
}

/*  This is defined in libvcd.h.  Just copied here so I can look at it more easily. 
typedef struct  _PsdPlayListDescriptor_tag {
  uint8 type;
  uint8 noi;  < number of items 
  uint16 lid; < list id: high-bit means this list is rejected in
                     the LOT (also, can't use 0) 
  uint16 prev_ofs;  previous list offset (0xffff disables) 
  uint16 next_ofs; < next list offset (0xffff disables)
  uint16 return_ofs; < return list offset (0xffff disables) 
  uint16 ptime; < play time in 1/15 s, 0x0000 meaning full item
  uint8  wtime; < delay after, in seconds, if 1 <= wtime <= 60 wait
                       is wtime else if 61 <= wtime <= 254 wait is
                       (wtime-60) * 10 + 60 else wtime == 255 wait is
                       infinite  
  uint8  atime; < auto pause wait time calculated same as wtime,
                       used for each item in list if the auto pause flag
                       in a sector is true 
  uint16 itemid[EMPTY_ARRAY_SIZE]; < item number
                                           0 <= n <= 1      - play nothing
                                           2 <= n <= 99     - play track n
                                         100 <= n <= 599    - play entry
                                                       (n - 99) from entries
                                                       table to end of track
                                         600 <= n <= 999    - reserved
                                        1000 <= n <= 2979   - play segment
                                                       play item (n - 999)
                                        2980 <= n <= 0xffff - reserved 
} GNUC_PACKED _PsdPlayListDescriptor_t;
*/
/* uint16 Pbc_PlayList (char *play_list)*/
static uint16 Pbc_PlayList(char *play_list)
{
int i; /* A multi-purpose integer. */
 uint16 prev_ofs, next_ofs, return_ofs;
 uint16 ptime;
 uint16 wtime, atime;
 uint16 *itemid;

 uint16 our_offset;
 uint64 then, now;
 int control_value = 0; 
   /* We can't convert the original copy to LE because we might be handed that
      PSD offset again, so we make a temporary copy of the 16-bit integers here
      and do our work based on those. */
      
   PlayList = (_PlayList *)play_list;
   itemid = (uint16 *)malloc(2*PlayList->noi);
   
   memcpy (itemid, play_list + 14, 2*PlayList->noi);
   prev_ofs =  PlayList->prev_ofs;
   next_ofs =  PlayList->next_ofs;
   return_ofs =  PlayList->return_ofs;
   ptime = PlayList->ptime;  /* Playtime in 1/15 seconds.  In order to support this, we need to convert to frames or something. */
   wtime = PlayList->wtime;  /* This one seems simple--wait (wtime) seconds before the next thing.  If 255, wait infinitely. */
   atime = PlayList->atime;  /* See TODO for where to find how to get the autopause flag from a sector. */
   
   /* Byteswap our copy. */
   prev_ofs = UINT16_SWAP_LE_BE_C(prev_ofs);
   next_ofs = UINT16_SWAP_LE_BE_C(next_ofs);
   return_ofs = UINT16_SWAP_LE_BE_C(return_ofs);
   ptime = UINT16_SWAP_LE_BE_C(ptime);
 
   for (i=0; i <  PlayList->noi; i++)
   {
    itemid[i] = UINT16_SWAP_LE_BE_C(itemid[i]);
   }
 
   our_offset = 0xffff;
   
   /* Wait time and Autopause time are in seconds if > 0 and <=60.  Otherwise, we have to recalculate them. */
    if ( (wtime > 60) && (wtime < 255) )
   {
       wtime = (wtime-60) * 10 + 60;
   }
   
   if ( (atime > 60) && (atime < 255) )
   {
       atime = (atime-60) * 10 + 60;
   } 
 
  /* Fill in our RemoteControl_t struct with the values for this list. */
   remotecontrol->bsn = 1; /* It doesn't matter in this case, but shouldn't ever be zero, really. */
   remotecontrol->nos = 0;
   remotecontrol->prev_ofs = prev_ofs;
   remotecontrol->next_ofs = next_ofs;
   remotecontrol->return_ofs = return_ofs;
   remotecontrol->default_ofs = 0xffff;
   remotecontrol->ptime = ptime; /* Playtime of item in 1/15 seconds. */
   remotecontrol->atime = atime; /* Autopause time in seconds. */
   remotecontrol->ExtList = NULL; /* No selection areas are present in PlayLists. */
   remotecontrol->jump_delay = 0;
/* Start a'Playin'! */ 
   for (i=0; i< PlayList->noi; i++)
   {
     if ( (itemid[i] >1)     && (itemid[i] < 100)  )  /* Play track i */
     { 
         control_value = Pbc_PlayTrack(itemid[i]);
     }else if ( (itemid[i] >=100)  && (itemid[i] < 600)  ) /* Play chapter/entry i-99 */
     {
     	 control_value = Pbc_PlayEntry(itemid[i]);	 
     }else if ( (itemid[i] >=1000)  && (itemid[i] < 2980)  ) /* Play Segment Item i-999 */
     {
     	 control_value = Pbc_PlaySpi(itemid[i]);
	 
     } /* else, we're kinda screwed because we've got an invalid itemid. */
     
     switch(control_value)
     {
     	case BUTTON_PREV:
        our_offset = prev_ofs;
        break;
     	case BUTTON_NEXT:
        our_offset = next_ofs;
        break;
     	case BUTTON_RET:
        our_offset = return_ofs;
        break;
	case BUTTON_STOP:
	our_offset = 0xffff;
	goto playlist_end;
	break;	 	 	 	 	 	 
    	default:
        break;
     }
     
     if (!(our_offset == 0xffff))
     {
       goto playlist_end;
     }
     then = now = timer_ms_gettime64();
     
     if (wtime > 0)
     {
        while ( (now - then)/1000 < wtime)
	{
	 now = timer_ms_gettime64();
	 control_value = vcd_dummy_poll();
	  switch(control_value)
     	  {
     	    case BUTTON_PREV:
            our_offset = prev_ofs;
            break;
     	    case BUTTON_NEXT:
            our_offset = next_ofs;
            break;
       	    case BUTTON_RET:
            our_offset = return_ofs;
            break;
	    case BUTTON_STOP:
	    our_offset = 0xffff;
	    goto playlist_end;
	    break;	 	 	 	 	 	 
    	    default:
            break;
     	  }
	 redraw_frame();
	 if ( !(our_offset == 0xffff) )
         {
          goto playlist_end;
         }
	 
	 if (PlayList->wtime == 255)
	 {
	  then = now;
	 }
	 
	}
	
	
	if ( !(our_offset == 0xffff) )
        {
          goto playlist_end;
        }
	
     }
     
     if ( !(our_offset == 0xffff) )
     {
       goto playlist_end;
     }
   }

playlist_end:
  if (our_offset == 0xffff) our_offset = next_ofs;  
  free (itemid); 
  return our_offset;
}

static void GenericEndPic(uint8 disc_no)
{
  pvr_wait_ready();
  pvr_scene_begin();
  pvr_list_begin(PVR_LIST_OP_POLY);
  /* Opaque list *************************************/
  bkg_render();
  /* End of opaque list */
  pvr_list_finish();
  pvr_list_begin(PVR_LIST_TR_POLY);
  /* Translucent list ********************************/
  draw_poly_strf(5.0f, 60.0f, 100.0f, 1.0f, 1.0f, 1.0f, 1.0f,
			VCDC_VERSION);//"              VC/DC  0.2  (C)2006-2009             ");
  draw_poly_strf(5.0f, 100.0f, 100.0f, 1.0f, 1.0f, 1.0f, 1.0f,
			"               Please Insert Disc # %d             ", disc_no);
			
  /* End of translucent list */
  pvr_list_finish();
  /* Finish the frame *******************************/
  pvr_scene_finish();
}
/*
typedef struct {
  uint8  type;         PSD_TYPE_END_LIST 
  uint8  next_disc;    0x00 to stop PBC or 0xnn to switch to disc no nn 
  uint16 change_pic;   0 or 1000..2979, should be still image 
  uint8  reserved[4];  padded with 0x00 
} GNUC_PACKED PsdEndListDescriptor_t;
*/
static uint16 Pbc_EndList(char *end_list)
{
  PsdEndListDescriptor_t *EndList;
  int redraw, control_value;
  
  EndList = (PsdEndListDescriptor_t *)end_list;
  
  remotecontrol->bsn = 1; /* It doesn't matter in this case, but shouldn't ever be zero, really. */
  remotecontrol->nos = 0;
  remotecontrol->prev_ofs = 0xffff;
  remotecontrol->next_ofs = 0xffff;
  remotecontrol->return_ofs = 0xffff;
  remotecontrol->default_ofs = 0xffff;
  remotecontrol->ptime = 0; /* Playtime of item in 1/15 seconds. */
  remotecontrol->atime = 0; /* Autopause time in seconds. */
  remotecontrol->ExtList = NULL; /* No selection areas are present in PlayLists. */
  remotecontrol->jump_delay = 0;
  
  if (EndList->next_disc)
  {
    
    if ( (EndList->change_pic >=1000)  && (EndList->change_pic < 2980)  ) /* Play Segment Item i-999 */
    {
     	 control_value = Pbc_PlaySpi(EndList->change_pic);
	 redraw = 1;	 
    } else {
    	 GenericEndPic(EndList->next_disc);
  	 redraw = 0;
    }
    while(1)
    {
      if (control_value == BUTTON_STOP) return 0xffff;
      control_value = vcd_dummy_poll();
      if (redraw)
      {
        redraw_frame();
      }else{
        GenericEndPic(EndList->next_disc);
      }
    }
  }
 return 0xffff;
}

/*void Fake_Pbc();
  This function will give us a 'good enough' simulation of Playback Control based on the contents of
  ENTRIES.VCD for discs without PBC available, or if user decides to turn off PBC in the menu.
  
  Basically, we can very simply compose enough information based on ENTRIES.VCD to trick our Pbc_PlayEntry(uint16 play_item)
  function into thinking we're in a PlayList or SelectionList.  Pretty neat, huh? If I figure out a fairly clean way to get
  our last sector position, this will allow for nicer chapter navigation than a VCD with PBC available.
  */

static int get_entry(int direction) /* We'll use 0 as "seek backward." */
{
  int new_entry;

  if (simple_entries->entry_count == 1) return 0; /* We'll end up here anyway... */   
  new_entry = simple_entries->entry_count-1;
  
  do {
    new_entry--;
  }  while (simple_entries->ent_lba[new_entry] > cur_sector);
  
  if (direction)
  {
   new_entry++;
  }else{
   new_entry--;
  }
  
  return new_entry;
}

void Fake_Pbc()
{
  int cur_entry;
  int disc_playing;
  int control_value;
  
  cur_entry = 0;
  control_value = 0;
  cur_sector = simple_entries->ent_lba[0];
  disc_playing = 1;
  
  bkg_clear(); /* The menu is using PVR memory that we need. */
  while (disc_playing==1)
  {
    /* Fill in our RemoteControl_t struct with the values for this list. */
   remotecontrol->bsn = 1;
   remotecontrol->nos = simple_entries->entry_count;
   remotecontrol->prev_ofs = (cur_entry > 0)? (cur_entry - 1) : 0xffff;
   remotecontrol->next_ofs = (cur_entry < simple_entries->entry_count-1) ? cur_entry+ 1 : 0xffff;
   remotecontrol->return_ofs = 0xffff;
   remotecontrol->default_ofs = 0xffff; 
   remotecontrol->ptime = 0x0000; /* Play full item--'ptime' is only specified in PlayList, not SelectionList. */
   remotecontrol->atime = 0; /* Autopause time not specified in SelectionList either. */
   remotecontrol->ExtList = NULL;
   remotecontrol->jump_delay = 0;
   control_value = Pbc_PlayEntry(cur_entry + 100); /* We need to add 100 since our program thinks we're in a SelectionList... */
   switch(control_value)
     {
     	case BUTTON_PREV:
        cur_entry = get_entry(0); /* Find out where we really are. */
        break;
     	case BUTTON_NEXT:
        cur_entry = get_entry(1);
        break;
     	case BUTTON_NUMERIC:
        cur_entry = pbc_get_button() - 1;
        break;
	case BUTTON_STOP:
	disc_playing = 0;
	break;	 	 	 	 	 	 
    	default:
	if (cur_sector <= simple_entries->ent_lba[simple_entries->entry_count - 1])
	{
	  cur_entry = get_entry(1);
	}else{
	  disc_playing = 0;
	}
        break;
     }
     
  }
  bkg_reset(); /* Put our menu graphics back now that we don't need the space any longer. */
    
}  

/*void Start_Pbc():
  This function is incomplete, but we've got the start of it--the first byte of the psdvcd array
  will tell us the type of Psd___ListDescriptor we have.  From there, we'll keep getting an offset.
  Multiply it by infovcd->offset_mult (maybe up above when we get the new offset) and get the byte at
  psdvcd[offset] for the type of the next Psd___ListDescriptor.  Once we know what kind of list we have,
  Do Pbc_PlayList or Pbc_SelectionList(psdvcd+offset); */
void Start_Pbc()
{
  int disc_playing;
  uint8 type;
  uint16 offset;

  if (!is_vcd())
  {
    return;
  }
  if ( (infovcd->psd_size == 0) || (pbc_switch == 0))
  {
    Fake_Pbc();
    return;
  }  
  type = psdvcd[0];
  
  offset = 0;
  disc_playing = 1;
  
  bkg_clear(); /* The menu is using PVR memory that we need. */
 
 while (disc_playing==1)
 {
  switch(type)
  {
    case PSD_TYPE_SELECTION_LIST:
    case PSD_TYPE_EXT_SELECTION_LIST:
    	offset = Pbc_SelectionList(psdvcd+(offset * infovcd->offset_mult));
    break;
    case PSD_TYPE_PLAY_LIST:
    	offset = Pbc_PlayList(psdvcd+(offset * infovcd->offset_mult));
    break;
    case PSD_TYPE_END_LIST:
      disc_playing = 0;
        /*Eh, show the "swap discs" picture if present (SVCD ONLY!), I guess.  Shouldn't be too hard. */
    break;
    case PSD_TYPE_COMMAND_LIST:
       disc_playing = 0;
    break; /* I have no friggin' clue what to do with the command list! It's only for SVCD...*/
    default:
    break;
  }
  if (offset == 0xffff)
  {
    disc_playing = 0;
  } else {
    type = psdvcd[offset * infovcd->offset_mult];
  }
 }
 bkg_reset(); /* Put our menu graphics back now that we don't need the space any longer. */ 
}

/* Checks if we have a VCD or not.  We should make use of this function to figure out what type--
  There is enough information in the first several bytes to tell if we have a VCD 1.1, VCD 2.0,
  SVCD, or HQ-VCD.  This will make a few things easier.  We can also set filenames here so that
  we can make the open_vcd() as simple as possible. */
#if 0
static int is_vcd()
{
		   
  if (
	(
	 infovcd->ID[0]== 'V' &&
	 infovcd->ID[1]== 'I' &&
	 infovcd->ID[2]== 'D' &&
	 infovcd->ID[3]== 'E' &&
	 infovcd->ID[4]== 'O' &&
	 infovcd->ID[5]== '_' &&
	 infovcd->ID[6]== 'C' &&
	 infovcd->ID[7]== 'D' ) ||
	(  
	 infovcd->ID[0]== 'S' &&
	 infovcd->ID[1]== 'U' &&
	 infovcd->ID[2]== 'P' &&
	 infovcd->ID[3]== 'E' &&
	 infovcd->ID[4]== 'R' &&
	 infovcd->ID[5]== 'V' &&
	 infovcd->ID[6]== 'C' &&
	 infovcd->ID[7]== 'D' ) ||
	(  
	 infovcd->ID[0]== 'H' &&
	 infovcd->ID[1]== 'Q' &&
	 infovcd->ID[2]== '-' &&
	 infovcd->ID[3]== 'V' &&
	 infovcd->ID[4]== 'C' &&
	 infovcd->ID[5]== 'D' &&
	 infovcd->ID[6]== ' ' &&
	 infovcd->ID[7]== ' ' )  
  ) /* We've got ourselves a (S)VCD! (I think.) */
  {
  	return 1;
  } /* Or not.*/
  return 0;
}
#else /* This is a bit better--tell which type of disc it is! */
static int is_vcd()
{
		   
  if 	(
	 infovcd->ID[0]== 'V' &&
	 infovcd->ID[1]== 'I' &&
	 infovcd->ID[2]== 'D' &&
	 infovcd->ID[3]== 'E' &&
	 infovcd->ID[4]== 'O' &&
	 infovcd->ID[5]== '_' &&
	 infovcd->ID[6]== 'C' &&
	 infovcd->ID[7]== 'D' )
	 {
	   strcpy(VCD_names.entries, "/vcd/VCD/ENTRIES.VCD");
	   strcpy(VCD_names.psd, "/vcd/VCD/PSD.VCD");
	   strcpy(VCD_names.lot, "/vcd/VCD/LOT.VCD");
	   
	   strcpy(VCD_names.psd_x, "/vcd/EXT/PSD_X.VCD");
	   strcpy(VCD_names.lot_x, "/vcd/EXT/LOT_X.VCD");
	   strcpy(VCD_names.scandata,"/vcd/EXT/SCANDATA.DAT");
	   return 1; /* We'll say '1' for VCD, '2' for SVCD/HQ-VCD. */
	 } 
  else if( (  
	 infovcd->ID[0]== 'S' &&
	 infovcd->ID[1]== 'U' &&
	 infovcd->ID[2]== 'P' &&
	 infovcd->ID[3]== 'E' &&
	 infovcd->ID[4]== 'R' &&
	 infovcd->ID[5]== 'V' &&
	 infovcd->ID[6]== 'C' &&
	 infovcd->ID[7]== 'D' ) ||
	 (  
	 infovcd->ID[0]== 'H' &&
	 infovcd->ID[1]== 'Q' &&
	 infovcd->ID[2]== '-' &&
	 infovcd->ID[3]== 'V' &&
	 infovcd->ID[4]== 'C' &&
	 infovcd->ID[5]== 'D' &&
	 infovcd->ID[6]== ' ' &&
	 infovcd->ID[7]== ' ' ) )
	 {
	   strcpy(VCD_names.entries, "/vcd/SVCD/ENTRIES.SVD");
	   strcpy(VCD_names.psd, "/vcd/SVCD/PSD.SVD");
	   strcpy(VCD_names.lot, "/vcd/SVCD/LOT.SVD");
	   strcpy(VCD_names.search, "/vcd/SVCD/SEARCH.DAT");
	   
	   /*  Extended PSD/LOT is incorporated into the standard PSD.SVD/LOT.SVD. */
	   strcpy(VCD_names.psd_x, "/vcd/SVCD/PSD.SVD");
	   strcpy(VCD_names.lot_x, "/vcd/SVCD/LOT.SVD");
	   strcpy(VCD_names.scandata,"/vcd/EXT/SCANDATA.DAT");
	   return 2;
	 }  
  
  return 0;
}
#endif

/* Copied in from KOS kernel/arch/dreamcast/hardware/cdrom.c */
#define MAKE_SYSCALL(rs, p1, p2, idx) \
	uint32 *syscall_bc = (uint32*)0x8c0000bc; \
	int (*syscall)() = (int (*)())(*syscall_bc); \
	rs syscall((p1), (p2), 0, (idx));

/* Set disc access mode */
static int gdc_change_data_type(void *param) { MAKE_SYSCALL(return, param, 0, 10); }

/* We want to get 2352 bytes for each sector we read every time we call cdrom_read_sectors. */
#if 0
static void read_mode_vcd()
{
  uint32 params[4];
  
   params[0] = 0;    /* 0 is set, 1 is get. */
   params[1] = 4096;  /* ?  4096 lets us read MPEG tracks
   			    on a VCD. */
   params[2] = 0x600; /*  ?  KOS uses 2048 if disc_type==0x20 or
   				1024 otherwise.  */
   params[3] = 2352;  /* sector size. */
		   
   gdc_change_data_type(params);
}		   
#endif

static int vcd_is_open = 0;
/*
  int open_vcd():
  Here, we allocate memory for the in-memory copies of certain vcd files, read those files into the buffers,
  and byteswap everything (stuff in here is Big-Endian, and we need things to be Little-Endian for the DC.
  On an unrelated note, I can't deal with Endian-ness issues without thinking about Gulliver's Travels by Jonathan Swift.
  Not that it helps me with this any, but I love Swift.
*/

int open_vcd()
{
  int i; /* A multi-purpose integer. */
  int fd; /* We'll be reading some files from track 0 of the VCD. */
  
  read_sectors = cdrom_read_sectors;
/* Allocate everything.  Check to make sure stuff isn't allocated already--if the user
   has changed discs, stuff will already be allocated! */
  if (infovcd == NULL) infovcd = (_InfoVcd_t *)malloc(2048);
  if (entriesvcd == NULL) entriesvcd = (_EntriesVcd_t *)malloc(2048);
  if (simple_entries == NULL) simple_entries = (_EntriesVcd_sim *)malloc(sizeof (_EntriesVcd_sim));
  
/* Start out with empty structures--we don't want junk data! */  
  memset(infovcd, 0, 2048);
  memset(entriesvcd, 0, 2048);
  memset(simple_entries, 0, sizeof(_EntriesVcd_sim));
  
/* Read INFO.VCD and ENTRIES.VCD. */
  cdrom_read_sectors(infovcd, 300, 1);
  if(!is_vcd()) return 0;  /* This is not a VCD! */
  
  /*cdrom_read_sectors(entriesvcd, 301, 1);  Handle this below... */ 
  
/* Byteswapping is needed here. We can probably ignore a lot of this anyway, but what the hell...*/
  infovcd->vol_count = UINT16_SWAP_LE_BE_C(infovcd->vol_count);
  infovcd->vol_id = UINT16_SWAP_LE_BE_C(infovcd->vol_id);
  infovcd->psd_size = UINT32_SWAP_LE_BE_C(infovcd->psd_size);
  MSF_TO_LBA(infovcd->first_seg_addr, first_seg_addr);
  infovcd->lot_entries = UINT16_SWAP_LE_BE_C(infovcd->lot_entries);
  infovcd->item_count = UINT16_SWAP_LE_BE_C(infovcd->item_count);
  for(i=0; i<5; i++)
  {
    infovcd->playing_time[i] = UINT16_SWAP_LE_BE_C(infovcd->playing_time[i]);
  }
  
  /*simplify_entries(entriesvcd, simple_entries);*/ /* This one is simple--converts ENTRIES.VCD into something we can deal with. */
#if 0  
  /* Move this below?  We want the extended versions if they exist, so it makes sense to check for those first. */
  /* Only allocate lotvcd and psdvcd if infovcd->psd_size > 0. */
  if (infovcd->psd_size > 0)
  {
    if (!(lotvcd==NULL))
    {
      free (lotvcd);
      lotvcd = NULL;
    }
    if (lotvcd==NULL) lotvcd = (_LotVcd_t *)malloc(2048*32);
    memset(lotvcd, 0, 2048*32);
    cdrom_read_sectors(lotvcd, 302, 32);
    for (i=0; i < LOT_VCD_OFFSETS; i++)
    {
     lotvcd->offset[i] = UINT16_SWAP_LE_BE_C(lotvcd->offset[i]);
    }
    /* Figure out the number of sectors PSD.VCD takes up -- it'll be psd_size/2048, or (psd_size/2048)+1; */
    i = infovcd->psd_size/2048;
    if (infovcd->psd_size % 2048) i++;
    if (!(psdvcd == NULL))
    {
      free (psdvcd);
      psdvcd = NULL;
    }
    if(psdvcd == NULL) psdvcd = malloc(i*2048);
    memset (psdvcd, 0, i*2048);
    cdrom_read_sectors(psdvcd, 334, i);
    
  } /*if (infovcd->psd_size > 0) */
#endif
  
 /*  read_mode_vcd(); */ /* Set us up to read 2352 bytes/sector. */

/* Initialize a modified version of the ISO-9660 filesystem that will read the non-typical VCD-type
   CD.  What this does is set the base track to track 0, rather than the last track with a control of
   40, because track 0 is were the actual filesystem on a VCD resides.  From here, we can find pointers to
   all of the files on the VCD, including tracks 1-99 (although I think I screwed up a tad, because playing
   those files as regular MPEG files seems to cause them to end early). For reading the files on track 0,
   this seems to work perfectly, exactly as one would expect when reading a file using the stock iso9660 fs
   stuff found in KOS. */

/* We don't want/need two CD-ROM filesystems trying to work at once, so shut down the stock ISO9660 filesystem. */

  fs_iso9660_shutdown();

/* Initialize our own hacked-up versin of the ISO9660 filesystem. */
  fs_isobridge_init();
  
  /* The fs_isobridge filesystem module will only reset the CD sector size after trying to open a file, so we
     will do some file reading.  We will start with ENTRIES.VCD, because it is a file that we'll always have. */

/* TODO!  Handle the SVCD equivalents of these filenames.  All but SCANDATA.DAT will have the same internal
  structure. */     
  /* fd = fs_open("/vcd/VCD/ENTRIES.VCD", O_RDONLY); */
   fd = fs_open(VCD_names.entries, O_RDONLY); 
  if (!(fd == -1))
  {
    i = fs_total(fd);
    fs_read(fd, entriesvcd, i);
    fs_close(fd);
    
    simplify_entries(entriesvcd, simple_entries);
  } else {
     close_vcd();
     return 0;
  }
  
  if (infovcd->psd_size > 0)
  {
   if  (infovcd->flags.pbc_x == true)
   {
    /* fd = fs_open ("/vcd/EXT/LOT_X.VCD", O_RDONLY); */
    fd = fs_open (VCD_names.lot_x, O_RDONLY);
    if (!(fd == -1))
    {
      if(!(lotvcd==NULL))
      {
        free(lotvcd);
      }
      i = fs_total(fd);
      lotvcd = (_LotVcd_t *)malloc(i);
      fs_read (fd, lotvcd, i);
      fs_close(fd);
      
      for (i=0; i < LOT_VCD_OFFSETS; i++)
      {
       lotvcd->offset[i] = UINT16_SWAP_LE_BE_C(lotvcd->offset[i]);
      }
    } 
    /* fd = fs_open("/vcd/ext/psd_x.vcd", O_RDONLY); */
    fd = fs_open (VCD_names.psd_x, O_RDONLY);
    if (!(fd == -1))
    {
      if(!(psdvcd==NULL))
      {
        free(psdvcd);
      }
      i = fs_total(fd);
      psdvcd = malloc(i);
      fs_read (fd, psdvcd, i);
      fs_close(fd);
    }
    
    /* SEARCH.DAT is almost the same--converting one type to the other should be easy! */
    /* Also, the SVCD SCANDATA.DAT contains the same information, so we should be able to
      figure out a way to convert any of these types into a simple type! */
    
    /* fd = fs_open("/vcd/EXT/SCANDATA.DAT", O_RDONLY); */
    fd = fs_open(VCD_names.scandata, O_RDONLY);
    if (!(fd == -1))
    {
      if (!(scandata==NULL))
      {
        free(scandata);
      }
      i = fs_total(fd);
      scandata = (ScandataDat_v2_t *)malloc(i);
      fs_read(fd, scandata, i);
      fs_close(fd);
    }
    
   }/* end of (infovcd->flags.pbc_x == true) ) */
   else
   {
    /* fd = fs_open ("/vcd/EXT/LOT.VCD", O_RDONLY); */
    fd = fs_open (VCD_names.lot, O_RDONLY);
    if (!(fd == -1))
    {
      if(!(lotvcd==NULL))
      {
        free(lotvcd);
      }
      i = fs_total(fd);
      lotvcd = (_LotVcd_t *)malloc(i);
      fs_read (fd, lotvcd, i);
      fs_close(fd);
      
      for (i=0; i < LOT_VCD_OFFSETS; i++)
      {
       lotvcd->offset[i] = UINT16_SWAP_LE_BE_C(lotvcd->offset[i]);
      }
    } 
    /* fd = fs_open("/vcd/ext/psd.vcd", O_RDONLY); */
    fd = fs_open (VCD_names.psd, O_RDONLY);
    if (!(fd == -1))
    {
      if(!(psdvcd==NULL))
      {
        free(psdvcd);
      }
      i = fs_total(fd);
      psdvcd = malloc(i);
      fs_read (fd, psdvcd, i);
      fs_close(fd);
    }
   } /* End of else */
  } /* end of if (infovcd->psd_size > 0) */
  if (remotecontrol == NULL)
  {
    remotecontrol = (RemoteControl_t *)malloc(sizeof(RemoteControl_t));
  }
  
  if (vcd_buffer==NULL)
  {
    vcd_buffer = (uint8_t *)malloc(2352*10);
  }
  
  if (vcd_toc==NULL)
  {
    vcd_toc = (CDROM_TOC *)malloc(sizeof(CDROM_TOC));
  }
  memset (vcd_toc, 0, sizeof(CDROM_TOC));
  read_toc = cdrom_read_toc;
  vcd_is_open = 1;
  return 1; 
}

void close_vcd()
{
 /* We don't want to close anything if we didn't open it already! */
 if (vcd_is_open){
  fs_isobridge_shutdown();
  fs_iso9660_init();
  
  /* Rather than free everything (the buffers don't take up all that much room)
     we will just set the ID string to spell VIDEO_CD backwards, which is invalid. */
  infovcd->ID[0]= 'D';
  infovcd->ID[1]= 'C';
  infovcd->ID[2]= '_';
  infovcd->ID[3]= 'O';
  infovcd->ID[4]= 'E';
  infovcd->ID[5]= 'D';
  infovcd->ID[6]= 'I';
  infovcd->ID[7]= 'V';
 }
 vcd_is_open = 0;
  
}

int play_vcd_image(const char* fn)
{
  int i; /* A multi-purpose integer. */
  int fd; /* We'll be reading some files from track 0 of the VCD. */
  
  i = fs_bincue_mount(fn);
  if (i == -1)
  {
    fs_bincue_unmount();
    return 0;
  }
  
  bincue_set_sector_size(2048); /* We want 2048 bytes/sector at first. */
  
/* Allocate everything.  Check to make sure stuff isn't allocated already--if the user
   has changed discs, stuff will already be allocated! */
  if (infovcd == NULL) infovcd = (_InfoVcd_t *)malloc(2048);
  if (entriesvcd == NULL) entriesvcd = (_EntriesVcd_t *)malloc(2048);
  if (simple_entries == NULL) simple_entries = (_EntriesVcd_sim *)malloc(sizeof (_EntriesVcd_sim));
  
/* Start out with empty structures--we don't want junk data! */  
  memset(infovcd, 0, 2048);
  memset(entriesvcd, 0, 2048);
  memset(simple_entries, 0, sizeof(_EntriesVcd_sim));
  
/* Read INFO.VCD and ENTRIES.VCD. */
  read_sectors(infovcd, 300, 1);
  if(!is_vcd())
  {
   fs_bincue_unmount();
   bsod();
   return 0;  /* This is not a VCD! */ 
  }
/* Byteswapping is needed here. We can probably ignore a lot of this anyway, but what the hell...*/
  infovcd->vol_count = UINT16_SWAP_LE_BE_C(infovcd->vol_count);
  infovcd->vol_id = UINT16_SWAP_LE_BE_C(infovcd->vol_id);
  infovcd->psd_size = UINT32_SWAP_LE_BE_C(infovcd->psd_size);
  MSF_TO_LBA(infovcd->first_seg_addr, first_seg_addr);
  infovcd->lot_entries = UINT16_SWAP_LE_BE_C(infovcd->lot_entries);
  infovcd->item_count = UINT16_SWAP_LE_BE_C(infovcd->item_count);
  for(i=0; i<5; i++)
  {
    infovcd->playing_time[i] = UINT16_SWAP_LE_BE_C(infovcd->playing_time[i]);
  }
  

 /* Set us up to read 2352 bytes/sector. */
   bincue_set_sector_size(2352);
  
/* TODO!  Handle the SVCD equivalents of these filenames.  All but SCANDATA.DAT will have the same internal
  structure. */
  
  fd = fs_open(VCD_names.entries, O_RDONLY); 
  if (!(fd == -1))
  {
    i = fs_total(fd);
    fs_read(fd, entriesvcd, i);
    fs_close(fd);
    
    simplify_entries(entriesvcd, simple_entries);
  } else {
     fs_bincue_unmount();
     bsod();
     return 0;
  }
  
  if (infovcd->psd_size > 0)
  {
   if  (infovcd->flags.pbc_x == true)
   {
    fd = fs_open (VCD_names.lot_x, O_RDONLY);
    if (!(fd == -1))
    {
      if(!(lotvcd==NULL))
      {
        free(lotvcd);
      }
      i = fs_total(fd);
      lotvcd = (_LotVcd_t *)malloc(i);
      fs_read (fd, lotvcd, i);
      fs_close(fd);
      
      for (i=0; i < LOT_VCD_OFFSETS; i++)
      {
       lotvcd->offset[i] = UINT16_SWAP_LE_BE_C(lotvcd->offset[i]);
      }
    } 
 
    fd = fs_open (VCD_names.psd_x, O_RDONLY);
    if (!(fd == -1))
    {
      if(!(psdvcd==NULL))
      {
        free(psdvcd);
      }
      i = fs_total(fd);
      psdvcd = malloc(i);
      fs_read (fd, psdvcd, i);
      fs_close(fd);
    }
    
    /* SEARCH.DAT is almost the same--converting one type to the other should be easy! */
    /* Also, the SVCD SCANDATA.DAT contains the same information, so we should be able to
      figure out a way to convert any of these types into a simple type! */
    
    fd = fs_open(VCD_names.scandata, O_RDONLY);
    if (!(fd == -1))
    {
      if (!(scandata==NULL))
      {
        free(scandata);
      }
      i = fs_total(fd);
      scandata = (ScandataDat_v2_t *)malloc(i);
      fs_read(fd, scandata, i);
      fs_close(fd);
    }
    
   }/* end of (infovcd->flags.pbc_x == true) ) */
   else
   {
    fd = fs_open (VCD_names.lot, O_RDONLY);
    if (!(fd == -1))
    {
      if(!(lotvcd==NULL))
      {
        free(lotvcd);
      }
      i = fs_total(fd);
      lotvcd = (_LotVcd_t *)malloc(i);
      fs_read (fd, lotvcd, i);
      fs_close(fd);
      
      for (i=0; i < LOT_VCD_OFFSETS; i++)
      {
       lotvcd->offset[i] = UINT16_SWAP_LE_BE_C(lotvcd->offset[i]);
      }
    } 
    fd = fs_open (VCD_names.psd, O_RDONLY);
    if (!(fd == -1))
    {
      if(!(psdvcd==NULL))
      {
        free(psdvcd);
      }
      i = fs_total(fd);
      psdvcd = malloc(i);
      fs_read (fd, psdvcd, i);
      fs_close(fd);
    }
   } /* End of else */
  } /* end of if (infovcd->psd_size > 0) */
  if (remotecontrol == NULL)
  {
    remotecontrol = (RemoteControl_t *)malloc(sizeof(RemoteControl_t));
  }
  
  if (vcd_buffer==NULL)
  {
    vcd_buffer = (uint8_t *)malloc(2352*10);
  }
  
  if (vcd_toc==NULL)
  {
    vcd_toc = (CDROM_TOC *)malloc(sizeof(CDROM_TOC));
  }
  memset (vcd_toc, 0, sizeof(CDROM_TOC));
  read_toc = vcd_read_toc;
  
  Start_Pbc();
  fs_bincue_unmount();
  read_toc = cdrom_read_toc;
    return 0; 
}
