/*
 * vidclock.h
 * Copyright (C) 2008 Josh Sutherland <OneThirty8@aol.com>
 * This file is part of VC/DC, a free MPEG-2 video player.
 *
 * VC/DC is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * VC/DC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * This is based on an idea I got from xine-lib, looking at code in
 * xine-engine/metronom.c and xine-engine/resample.c
 * Aditionally, much code is taken from mpeg2dec by Michel Lespinasse
 * and Aaron Holtzman, particularly the print_fps (int final); function, which
 * is reused in vidclock.c to check whether we're running fast or slow.
 */
  
#include <kos.h>

#include "inttypes.h"
#include "mpeg2.h"
#include "gettimeofday.h"

/* Initialize the a/v synch clock. */
void vidclock_init(unsigned int frame_period);

/* This is good enough for video-only sequences, but for streams with audio we can
   probably be even more accurate by looking at the number of samples that have played. */
int vidclock_get_ptime();

/* Reset the a/v synch clock, for example if we have paused or fast-forwarded. */
void vidclock_reset();

/* Set the sample rate that the video clock should use for synchronizing the audio. */
void vidclock_set_audio_rate(int freqs, int chans, int buf_size);

/* Get a minimum threshold below which we won't bother resampling. Your program will have to handle such a case--
  this just returns zero or the number of samples that play per half a video frame (roughly). */
int vidclock_drift_allowance();

/* We should only call this once per call to the audio decoder function. Also, rather than synch audio to video,
  we'll probably do better to synch to an external clock. We just need to figure a way to keep the clock from going
  wonky if the disc has spun down during a pause or something. */
int vidclock_handle_audio(uint32_t samples);

/* Tell the video clock that we have just been given a new frame of video to decode, and it will decide if
  we should decode it or skip it. */
void vidclock_handle_incoming (const mpeg2_info_t * info, mpeg2dec_t *mpeg2dec);    		

/* Tell the video clock that we have just been told that a skipped frame is ready--obviously we can't display
  a non-existant frame, so we pretend that we have and go on our merry way. */
void vidclock_skip_frame();

/* Tell the video clock that we have just been told that a frame of decoded video is ready for display. This
  function should be called from within a separate thread from the main program, as it will wait for it to be
  time to display the current frame before returning.  Running in the same thread as the decoders will cause
  your program to run too slowly. */
void vidclock_handle_frame ();

